"""Classifiers used in the publication."""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]

from tsml_eval.utils.functions import str_in_nested_list

bakeoff_classifiers = [
    # distance based
    ["KNeighborsTimeSeriesClassifier", "dtw", "1nn-dtw"],
    ["GRAILClassifier", "grail"],
    # feature based
    ["Catch22Classifier", "catch22"],
    ["FreshPRINCEClassifier", "freshprince"],
    ["TSFreshClassifier", "tsfresh"],
    ["SignatureClassifier", "signatures"],
    # shapelet based
    ["ShapeletTransformClassifier", "stc", "stc-2hour"],
    ["RDSTClassifier", "rdst"],
    ["RandomShapeletForestClassifier", "randomshapeletforest", "rsf"],
    ["MrSQMClassifier", "mrsqm"],
    # interval based
    ["RSTSFClassifier", "rstsf", "r-stsf"],
    ["RandomIntervalSpectralEnsembleClassifier", "rise"],
    ["TimeSeriesForestClassifier", "tsf"],
    ["CanonicalIntervalForestClassifier", "cif"],
    ["SupervisedTimeSeriesForest", "stsf"],
    ["drcif", "DrCIFClassifier"],
    ["quant", "QUANTClassifier"],
    # dictionary based
    ["BOSSEnsemble", "boss"],
    ["ContractableBOSS", "cboss"],
    ["TemporalDictionaryEnsemble", "tde"],
    ["WEASEL", "weasel v1", "weasel v1.0"],
    ["weasel_v2", "weasel v2", "weasel v2.0", "weasel-dilation", "weasel-d"],
    # convolution based
    ["RocketClassifier", "rocket"],
    ["minirocket", "mini-rocket"],
    ["multirocket", "multi-rocket"],
    ["arsenalclassifier", "Arsenal"],
    ["hydra", "hydraclassifier"],
    ["mr-hydra", "multirockethydra", "multirockethydraclassifier", "multirocket-hydra"],
    # deep learning
    ["CNNClassifier", "cnn"],
    ["ResNetClassifier", "resnet"],
    ["InceptionTimeClassifier", "inceptiontime"],
    ["h-inceptiontimeclassifier", "h-inceptiontime"],
    ["LITETimeClassifier", "litetime"],
    # hybrid
    ["HIVECOTEV1", "hc1"],
    ["HIVECOTEV2", "hc2"],
    ["RIST", "RISTClassifier"],
]


def _set_bakeoff_classifier(
    classifier_name,
    random_state=None,
    n_jobs=1,
    **kwargs,
):
    c = classifier_name.lower()

    if not str_in_nested_list(bakeoff_classifiers, c):
        raise ValueError(f"UNKNOWN CLASSIFIER: {c} in _set_bakeoff_classifier")

    if c == "kneighborstimeseriesclassifier" or c == "dtw" or c == "1nn-dtw":
        from aeon.classification.distance_based import KNeighborsTimeSeriesClassifier

        return KNeighborsTimeSeriesClassifier(
            distance="dtw", n_neighbors=1, n_jobs=n_jobs, **kwargs
        )
    elif c == "grailclassifier" or c == "grail":
        from tsml.distance_based import GRAILClassifier

        return GRAILClassifier(**kwargs)
    elif c == "catch22classifier" or c == "catch22":
        from aeon.classification.feature_based import Catch22Classifier
        from sklearn.ensemble import RandomForestClassifier

        return Catch22Classifier(
            estimator=RandomForestClassifier(n_estimators=500),
            catch24=False,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "freshprinceclassifier" or c == "freshprince":
        from aeon.classification.feature_based import FreshPRINCEClassifier

        return FreshPRINCEClassifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "tsfreshclassifier" or c == "tsfresh":
        from aeon.classification.feature_based import TSFreshClassifier

        return TSFreshClassifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "signatureclassifier" or c == "signatures":
        from aeon.classification.feature_based import SignatureClassifier

        return SignatureClassifier(
            random_state=random_state,
            **kwargs,
        )
    elif c == "shapelettransformclassifier" or c == "stc" or c == "stc-2hour":
        from aeon.classification.shapelet_based import ShapeletTransformClassifier

        return ShapeletTransformClassifier(
            n_shapelet_samples=10000,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "rdstclassifier" or c == "rdst":
        from aeon.classification.shapelet_based import RDSTClassifier

        return RDSTClassifier(
            random_state=random_state,
            **kwargs,
        )
    elif (
        c == "randomshapeletforestclassifier"
        or c == "randomshapeletforest"
        or c == "rsf"
    ):
        from tsml.shapelet_based import RandomShapeletForestClassifier

        return RandomShapeletForestClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "mrsqmclassifier" or c == "mrsqm":
        from aeon.classification.shapelet_based import MrSQMClassifier

        return MrSQMClassifier(
            random_state=random_state,
            **kwargs,
        )
    elif c == c == "rstsfclassifier" or c == "rstsf" or c == "r-stsf":
        from tsml.interval_based import RSTSFClassifier

        return RSTSFClassifier(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "randomintervalspectralensembleclassifier" or c == "rise":
        from aeon.classification.interval_based import (
            RandomIntervalSpectralEnsembleClassifier,
        )

        return RandomIntervalSpectralEnsembleClassifier(
            n_estimators=500,
            min_interval_length=16,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "timeseriesforestclassifier" or c == "tsf":
        from aeon.classification.interval_based import TimeSeriesForestClassifier

        return TimeSeriesForestClassifier(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "canonicalintervalforestclassifier" or c == "cif":
        from aeon.classification.interval_based import CanonicalIntervalForestClassifier

        return CanonicalIntervalForestClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "supervisedtimeseriesforest" or c == "stsf":
        from aeon.classification.interval_based import SupervisedTimeSeriesForest

        return SupervisedTimeSeriesForest(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "drcif" or c == "drcifclassifier":
        from aeon.classification.interval_based import DrCIFClassifier

        return DrCIFClassifier(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "quantclassifier" or c == "quant":
        from aeon.classification.interval_based import QUANTClassifier

        return QUANTClassifier(random_state=random_state, **kwargs)
    elif c == "bossensemble" or c == "boss":
        from aeon.classification.dictionary_based import BOSSEnsemble

        return BOSSEnsemble(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "contractableboss" or c == "cboss":
        from aeon.classification.dictionary_based import ContractableBOSS

        return ContractableBOSS(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "temporaldictionaryensemble" or c == "tde":
        from aeon.classification.dictionary_based import TemporalDictionaryEnsemble

        return TemporalDictionaryEnsemble(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "weasel" or c == "weasel v1" or c == "weasel v1.0":
        from aeon.classification.dictionary_based import WEASEL

        return WEASEL(
            random_state=random_state,
            n_jobs=n_jobs,
            support_probabilities=True,
            **kwargs,
        )
    elif (
        c == "weasel_v2"
        or c == "weasel v2"
        or c == "weasel v2.0"
        or c == "weasel-dilation"
        or c == "weasel-d"
    ):
        from aeon.classification.dictionary_based import WEASEL_V2

        return WEASEL_V2(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "rocketclassifier" or c == "rocket":
        from aeon.classification.convolution_based import RocketClassifier

        return RocketClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "minirocket" or c == "mini-rocket":
        from aeon.classification.convolution_based import RocketClassifier

        return RocketClassifier(
            rocket_transform="minirocket",
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "multirocket" or c == "multi-rocket":
        from aeon.classification.convolution_based import RocketClassifier

        return RocketClassifier(
            rocket_transform="multirocket",
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "arsenalclassifier" or c == "arsenal":
        from aeon.classification.convolution_based import Arsenal

        return Arsenal(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "hydra" or c == "hydraclassifier":
        from aeon.classification.convolution_based import HydraClassifier

        return HydraClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif (
        c == "mr-hydra"
        or c == "multirockethydra"
        or c == "multirockethydraclassifier"
        or c == "multirocket-hydra"
    ):
        from aeon.classification.convolution_based import MultiRocketHydraClassifier

        return MultiRocketHydraClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "cnnclassifier" or c == "cnn":
        from aeon.classification.deep_learning import CNNClassifier

        return CNNClassifier(
            random_state=random_state,
            **kwargs,
        )
    elif c == "resnetclassifier" or c == "resnet":
        from aeon.classification.deep_learning import ResNetClassifier

        return ResNetClassifier(
            random_state=random_state,
            **kwargs,
        )
    elif c == "inceptiontimeclassifier" or c == "inceptiontime":
        from aeon.classification.deep_learning import InceptionTimeClassifier

        return InceptionTimeClassifier(
            random_state=random_state,
            **kwargs,
        )
    elif c == "h-inceptiontimeclassifier" or c == "h-inceptiontime":
        from aeon.classification.deep_learning import InceptionTimeClassifier

        return InceptionTimeClassifier(
            use_custom_filters=True, random_state=random_state, **kwargs
        )
    elif c == "litetimeclassifier" or c == "litetime":
        from aeon.classification.deep_learning import LITETimeClassifier

        return LITETimeClassifier(random_state=random_state, **kwargs)
    elif c == "hivecotev1" or c == "hc1":
        from aeon.classification.hybrid import HIVECOTEV1

        return HIVECOTEV1(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "hivecotev2" or c == "hc2":
        from aeon.classification.hybrid import HIVECOTEV2

        return HIVECOTEV2(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "ristclassifier" or c == "rist":
        from aeon.classification.hybrid import RISTClassifier
        from sklearn.ensemble import ExtraTreesClassifier

        return RISTClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=ExtraTreesClassifier(n_estimators=500, criterion="entropy"),
            **kwargs,
        )



"""Tests for the results validation utilities."""

import pytest

from tsml_eval.testing.testing_utils import _TEST_RESULTS_PATH
from tsml_eval.utils.results_validation import validate_results_file


@pytest.mark.parametrize(
    "path",
    [
        _TEST_RESULTS_PATH + "/regression/regressionResultsFile.csv",
        _TEST_RESULTS_PATH + "/classification/classificationResultsFile1.csv",
    ],
)
def test_validate_results_file(path):
    """Test results file validation with valid files."""
    assert validate_results_file(path)


@pytest.mark.parametrize(
    "path",
    [
        _TEST_RESULTS_PATH + "/broken/brokenRegressionResultsFile.csv",
        _TEST_RESULTS_PATH + "/broken/brokenClassificationResultsFile.csv",
        _TEST_RESULTS_PATH + "/broken/brokenResultsFile.csv",
    ],
)
def test_validate_broken_results_file(path):
    """Test results file validation with broken files."""
    assert not validate_results_file(path)



"""Simple vote ensemble for clustering."""

import numpy as np
import pandas as pd
from scipy.optimize import linear_sum_assignment
from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.cluster import KMeans
from sklearn.utils import check_random_state
from tsml.base import _clone_estimator


class SimpleVote(BaseEstimator, ClusterMixin):
    """
    SimpleVote clustering ensemble.

    SimpleVote is a clustering ensemble that uses a voting scheme to combine
    the results of multiple clusterers.

    Parameters
    ----------
    clusterers : list of clusterers, default=None
        A list of clusterers to use in the ensemble. If None, defaults to 5
        KMeans clusterers.
    n_clusters : int, default=8
        The number of clusters to form.
    random_state : int, default=None
        The seed for random number generation.

    Attributes
    ----------
    labels_ : ndarray of shape (n_instances,)
        Labels of each point from the last fit.

    Examples
    --------
    >>> from tsml_eval.estimators.clustering.consensus.simple_vote import SimpleVote
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.metrics import rand_score
    >>> iris = load_iris()
    >>> sv = SimpleVote(n_clusters=3, random_state=0)
    >>> sv.fit(iris.data)
    SimpleVote(...)
    >>> rand_score(iris.target, sv.labels_)
    0.8737360178970918
    """

    def __init__(self, clusterers=None, n_clusters=8, random_state=None):
        self.clusterers = clusterers
        self.n_clusters = n_clusters
        self.random_state = random_state

    def fit(self, X, y=None):
        """Fit model to X using a simple vote ensemble."""
        if isinstance(X, np.ndarray) and len(X.shape) == 3 and X.shape[1] == 1:
            X = np.reshape(X, (X.shape[0], -1))
        elif isinstance(X, pd.DataFrame) and len(X.shape) == 2:
            X = X.to_numpy()
        elif isinstance(X, list):
            X = np.array(X)
        elif not isinstance(X, np.ndarray) or len(X.shape) > 2:
            raise ValueError(
                "SimpleVote is not a time series clusterer. "
                "A valid sklearn input such as a 2d numpy array is required."
                "Sparse input formats are currently not supported."
            )
        X = self._validate_data(X=X, ensure_min_samples=self.n_clusters)

        rng = check_random_state(self.random_state)

        if self.clusterers is None:
            self._clusterers = [
                KMeans(
                    n_clusters=self.n_clusters,
                    n_init="auto",
                    random_state=rng.randint(np.iinfo(np.int32).max),
                )
                for _ in range(5)
            ]
        else:
            self._clusterers = [
                _clone_estimator(
                    clusterer, random_state=rng.randint(np.iinfo(np.int32).max)
                )
                for clusterer in self.clusterers
            ]

        cluster_assignments = np.zeros(
            (len(self._clusterers), X.shape[0]), dtype=np.int32
        )
        for i, clusterer in enumerate(self._clusterers):
            clusterer.fit(X)
            cluster_assignments[i] = clusterer.labels_

            uc = np.unique(clusterer.labels_)
            if uc.shape[0] != self.n_clusters:
                raise ValueError(
                    "Input clusterers must have the same number of clusters as the "
                    f"SimpleVote n_clusters ({self.n_clusters}). Found "
                    f"{uc.shape[0]} for clusterer {i}."
                )
            elif (np.sort(uc) != np.arange(self.n_clusters)).any():
                raise ValueError(
                    "Input clusterers must have cluster labels in the range "
                    "0 to  n_clusters - 1."
                )

        self._build_ensemble(cluster_assignments)

        return self

    def predict(self, X):
        """Predict cluster labels for X."""
        rng = check_random_state(self.random_state)
        return np.array(
            [
                rng.choice(np.flatnonzero(prob == prob.max()))
                for prob in self.predict_proba(X)
            ]
        )

    def predict_proba(self, X):
        """Predict cluster probabilities for X."""
        if isinstance(X, np.ndarray) and len(X.shape) == 3 and X.shape[1] == 1:
            X = np.reshape(X, (X.shape[0], -1))
        elif isinstance(X, pd.DataFrame) and len(X.shape) == 2:
            X = X.to_numpy()
        elif isinstance(X, list):
            X = np.array(X)
        elif not isinstance(X, np.ndarray) or len(X.shape) > 2:
            raise ValueError(
                "SimpleVote is not a time series clusterer. "
                "A valid sklearn input such as a 2d numpy array is required."
                "Sparse input formats are currently not supported."
            )
        X = self._validate_data(X=X, reset=False)

        cluster_assignments = np.zeros(
            (len(self._clusterers), X.shape[0]), dtype=np.int32
        )

        cluster_assignments[0] = self._clusterers[0].predict(X)
        for i in range(1, len(self._clusterers)):
            cluster_assignments[i] = self._new_labels[i - 1][
                self._clusterers[i].predict(X)
            ]

        votes = np.apply_along_axis(
            lambda x: np.bincount(x, minlength=self.n_clusters),
            axis=0,
            arr=cluster_assignments,
        ).T

        return votes / len(self._clusterers)

    def _build_ensemble(self, cluster_assignments):
        rng = check_random_state(self.random_state)

        self._new_labels = []
        new_assignments = np.zeros(
            (cluster_assignments.shape[0], cluster_assignments.shape[1]), dtype=np.int32
        )

        new_assignments[0] = cluster_assignments[0]
        for i in range(1, len(cluster_assignments)):
            cost_matrix = -np.histogram2d(
                cluster_assignments[i], cluster_assignments[0], bins=self.n_clusters
            )[0]
            _, col_indices = linear_sum_assignment(cost_matrix)
            self._new_labels.append(col_indices)
            new_assignments[i] = col_indices[cluster_assignments[i]]

        votes = np.apply_along_axis(
            lambda x: np.bincount(x, minlength=self.n_clusters),
            axis=0,
            arr=new_assignments,
        ).T

        self.labels_ = np.array(
            [rng.choice(np.flatnonzero(v == v.max())) for v in votes]
        )



"""Experiment runner for bakeoff redux 2023 publication."""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]

import os
import sys

os.environ["MKL_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["NUMEXPR_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["OMP_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["TF_NUM_INTEROP_THREADS"] = "1"
os.environ["TF_NUM_INTRAOP_THREADS"] = "1"

from tsml.base import _clone_estimator

from tsml_eval.experiments import load_and_run_classification_experiment
from tsml_eval.publications.y2023.tsc_bakeoff.set_bakeoff_classifier import (
    _set_bakeoff_classifier,
)
from tsml_eval.publications.y2023.tsc_bakeoff.tests import _BAKEOFF_TEST_RESULTS_PATH
from tsml_eval.testing.testing_utils import _TEST_DATA_PATH
from tsml_eval.utils.arguments import parse_args
from tsml_eval.utils.experiments import _results_present

# all classifiers ran without duplicates
distance_based = ["1NN-DTW", "GRAIL"]
feature_based = ["Catch22", "FreshPRINCE", "TSFresh", "Signatures"]
shapelet_based = ["STC", "RDST", "RSF", "MrSQM"]
interval_based = ["R-STSF", "RISE", "TSF", "CIF", "STSF", "DrCIF", "QUANT"]
dictionary_based = ["BOSS", "cBOSS", "TDE", "WEASEL", "WEASEL-D"]
convolution_based = [
    "ROCKET",
    "MiniROCKET",
    "MultiROCKET",
    "Arsenal",
    "Hydra",
    "Hydra-MultiROCKET",
]
deep_learning = ["CNN", "ResNet", "InceptionTime", "H-InceptionTime", "LITETime"]
hybrid = ["HC1", "HC2", "RIST"]
# top performing classifiers
top_classifiers = [
    "FreshPRINCE",
    "RDST",
    "QUANT",
    "WEASEL-D",
    "Hydra-MultiROCKET",
    "H-InceptionTime",
    "HC2",
]


def _run_experiment(args, predefined_resample):
    if args is None or args.__len__() < 1:
        data_path = _TEST_DATA_PATH
        results_path = _BAKEOFF_TEST_RESULTS_PATH
        classifier = "ROCKET"
        dataset_name = "MinimalChinatown"
        resample_id = 0
        n_jobs = 1
        kwargs = {}
        overwrite = False
    else:
        print("Input args = ", args)
        args = parse_args(args)
        data_path = args.data_path
        results_path = args.results_path
        classifier = args.estimator_name
        dataset_name = args.dataset_name
        resample_id = args.resample_id
        n_jobs = args.n_jobs
        kwargs = args.kwargs
        overwrite = args.overwrite
        predefined_resample = predefined_resample or args.predefined_resample

    # Skip if not overwrite and results already present
    # this is also checked in load_and_run, but doing a quick check here so can
    # print a message and make sure data is not loaded
    if not overwrite and _results_present(
        results_path,
        classifier,
        dataset_name,
        resample_id=resample_id,
        split="TEST",
    ):
        print("Ignoring, results already present")
    else:
        load_and_run_classification_experiment(
            data_path,
            results_path,
            dataset_name,
            (
                _set_bakeoff_classifier(
                    classifier,
                    random_state=resample_id,
                    n_jobs=n_jobs,
                    **kwargs,
                )
                if isinstance(classifier, str)
                else _clone_estimator(classifier, resample_id)
            ),
            classifier_name=classifier,
            resample_id=resample_id,
            overwrite=overwrite,
            predefined_resample=predefined_resample,
        )


if __name__ == "__main__":
    """
    Example simple usage, with arguments input via script or hard coded for testing.

    1. Edit the arguments from line 49-56 to suit your experiment, The most important
       are:
         data_path: the path to the data
         results_path: the path to the results
         classifier: the name of the classifier to use (check
         set_bakeoff_classifier.py), or an estimator object
         resample_id: the data resample id and random seed to use
    2. Run the script, if the experiment runs successfully a set of folders and a
       results csv file will be created in the results path.

    For evaluation of the written results, you can use the evaluation package, see
    our examples for usage:
    https://github.com/time-series-machine-learning/tsml-eval/blob/main/examples/

    For using your own classifier, any classifier following the sklearn, aeon,
    or tsml interface should be compatible with this file.
    """
    print("Running run_experiments.py main")
    # The for the UCR 112 datasets set this to true to exactly reproduce results in the
    # paper. This uses the randomly generated resamples from tsml-java if the file is
    # present (see notebook for link). Except for PF, the 30 new datasets use
    # tsml-eval resamples.
    predefined_resample = False
    _run_experiment(sys.argv[1:], predefined_resample)



"""Utility for recording the maximum memory usage of a function."""

import time
from threading import Thread

import psutil


def record_max_memory(
    function, args=None, kwargs=None, interval=0.1, return_func_time=False
):
    """
    Record the maximum memory usage of a function.

    Parameters
    ----------
    function : function
        The function to run.
    args : list, default=None
        The arguments to pass to the function.
    kwargs : dict, default=None
        The keyword arguments to pass to the function.
    interval : float, default=0.1
        The interval (in seconds) to check the memory usage.
    return_func_time : bool, default=False
        Whether to return the function's runtime.

    Returns
    -------
    max_memory : int
        The maximum memory usage (in bytes).
    runtime : int, optional
        The function's runtime (in milliseconds).

    Examples
    --------
    >>> def f(n):
    ...     return [i for i in range(n)]
    >>> max_mem = record_max_memory(f, args=[10000])
    """
    process = psutil.Process()
    start_memory = process.memory_info().rss

    thread = _FunctionThread(function, args, kwargs)
    thread.start()

    max_memory = process.memory_info().rss

    while True:
        time.sleep(interval)

        mem = process.memory_info().rss
        if mem > max_memory:
            max_memory = mem

        if not thread.is_alive():
            if thread.exception is not None:
                raise thread.exception

            if return_func_time:
                return max_memory - start_memory, thread.function_time
            else:
                return max_memory - start_memory


class _FunctionThread(Thread):
    """Thread that runs a function with arguments."""

    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}

        self.function_time = -1
        self.exception = None

        super().__init__(daemon=True)

    def run(self):
        """Overloads the threading.Thread.run."""
        start = int(round(time.time() * 1000))
        try:
            self.function(*self.args, **self.kwargs)
        except Exception as e:
            self.exception = e
        end = int(round(time.time() * 1000))
        self.function_time = end - start



"""tsml-eval command line argument parser."""

__author__ = ["MatthewMiddlehurst"]

__all__ = [
    "parse_args",
]

import argparse

import tsml_eval


def parse_args(args):
    """Parse the command line arguments for tsml_eval.

    The following is the --help output for tsml_eval:

    usage: tsml_eval [-h] [--version] [-ow] [-pr] [-rs RANDOM_SEED] [-nj N_JOBS]
                     [-tr] [-te] [-fc FIT_CONTRACT] [-ch] [-rn] [-nc N_CLUSTERS]
                     [-kw KEY VALUE TYPE]
                     data_path results_path estimator_name dataset_name
                     resample_id

    positional arguments:
      data_path             the path to the directory storing dataset files.
      results_path          the path to the directory where results files are
                            written to.
      estimator_name        the name of the estimator to run. See the
                            set_{task}.py file for each task learning task for
                            available options.
      dataset_name          the name of the dataset to load.
                            {data_dir}/{dataset_name}/{dataset_name}_TRAIN.ts and
                            {data_dir}/{dataset_name}/{dataset_name}_TEST.ts will
                            be loaded.
      resample_id           the resample ID to use when randomly resampling the
                            data, as a random seed for estimators and the suffix
                            when writing results files. An ID of 0 will use the
                            default TRAIN/TEST split.

    options:
      -h, --help            show this help message and exit
      --version             show program's version number and exit
      -ow, --overwrite      overwrite existing results files. If False, existing
                            results files will be skipped (default: False).
      -pr, --predefined_resample
                            load a dataset file with a predefined resample. The
                            dataset file must follow the naming format
                            '{dataset_name}_{resample_id}.ts' (default: False).
      -rs RANDOM_SEED, --random_seed RANDOM_SEED
                            use a different random seed than the resample ID. If
                            None use the {resample_id} (default: None).
      -nj N_JOBS, --n_jobs N_JOBS
                            the number of jobs to run in parallel. Only used if
                            the experiments file and selected estimator allows
                            threading (default: 1).
      -tr, --train_fold     write a results file for the training data in the
                            classification and regression task (default: False).
      -te, --test_fold      write a results file for the test data in the
                            clustering task (default: False).
      -fc FIT_CONTRACT, --fit_contract FIT_CONTRACT
                            a time limit for estimator fit in minutes. Only used
                            if the estimator can contract fit (default: 0).
      -ch, --checkpoint     save the estimator fit to file periodically while
                            building. Only used if the estimator can checkpoint
                            (default: False).
      -rn, --row_normalise  normalise the data rows prior to fitting and
                            predicting. (default: False).
      -nc N_CLUSTERS, --n_clusters N_CLUSTERS
                            the number of clusters to find for clusterers which
                            have an {n_clusters} parameter. If {-1}, use the
                            number of classes in the dataset (default: -1).
      -ctts, --combine_test_train_split
                            whether to use a train/test split or not. If True, the
                            train and test sets are combined and used the fit the
                            estimator. Only available for clustering
                            (default: False).
      -bt, --benchmark_time
                            run a benchmark function and save the time spent in the
                            results file (default: False).
      -wa, --write_attributes
                            write the estimator attributes to file when running
                            experiments. Will recursively write the attributes of
                            sub-estimators if present. (default: False).
      -ams ATT_MAX_SHAPE, --att_max_shape ATT_MAX_SHAPE
                            The max estimator collections shape allowed when
                            writing attributes, at 0 no estimators in collections
                            will be written, at 1 estimators in one-dimensional
                            lists will be written etc. (default: 0).
      -kw KEY VALUE TYPE, --kwargs KEY VALUE TYPE, --kwarg KEY VALUE TYPE
                            additional keyword arguments to pass to the estimator.
                            Should contain the parameter to set, the parameter
                            value, and the type of the value i.e. {--kwargs
                            n_estimators 200 int} to change the size of an
                            ensemble. Valid types are {int, float, bool, str}. Any
                            other type will be passed as a str. Can be used
                            multiple times (default: None).

    Parameters
    ----------
    args : list
        List of command line arguments to parse.

    Returns
    -------
    same_resample : argparse.Namespace
        The parsed command line arguments.
    """
    parser = argparse.ArgumentParser(prog="tsml_eval")
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {tsml_eval.__version__}"
    )
    parser.add_argument(
        "data_path", help="the path to the directory storing dataset files."
    )
    parser.add_argument(
        "results_path",
        help="the path to the directory where results files are written to.",
    )
    parser.add_argument(
        "estimator_name",
        help="the name of the estimator to run. See the set_{task}.py file for each "
        "task learning task for available options.",
    )
    parser.add_argument(
        "dataset_name",
        help="the name of the dataset to load. "
        "{data_dir}/{dataset_name}/{dataset_name}_TRAIN.ts and "
        "{data_dir}/{dataset_name}/{dataset_name}_TEST.ts will be loaded.",
    )
    parser.add_argument(
        "resample_id",
        type=int,
        help="the resample ID to use when randomly resampling the data, as a random "
        "seed for estimators and the suffix when writing results files. An ID of "
        "0 will use the default TRAIN/TEST split.",
    )
    parser.add_argument(
        "-ow",
        "--overwrite",
        action="store_true",
        help="overwrite existing results files. If False, existing results files "
        "will be skipped (default: %(default)s).",
    )
    parser.add_argument(
        "-pr",
        "--predefined_resample",
        action="store_true",
        help="load a dataset file with a predefined resample. The dataset file must "
        "follow the naming format '{dataset_name}{resample_id}.ts' "
        "(default: %(default)s).",
    )
    parser.add_argument(
        "-rs",
        "--random_seed",
        type=int,
        help="use a different random seed than the resample ID. If None use the "
        "{resample_id} (default: %(default)s).",
    )
    parser.add_argument(
        "-nj",
        "--n_jobs",
        type=int,
        default=1,
        help="the number of jobs to run in parallel. Only used if the experiments file "
        "and selected estimator allows threading (default: %(default)s).",
    )
    parser.add_argument(
        "-tr",
        "--train_fold",
        action="store_true",
        help="write a results file for the training data in the classification and "
        "regression task (default: %(default)s).",
    )
    parser.add_argument(
        "-te",
        "--test_fold",
        action="store_true",
        help="write a results file for the test data in the clustering task "
        "(default: %(default)s).",
    )
    parser.add_argument(
        "-fc",
        "--fit_contract",
        type=int,
        default=0,
        help="a time limit for estimator fit in minutes. Only used if the estimator "
        "can contract fit (default: %(default)s).",
    )
    parser.add_argument(
        "-ch",
        "--checkpoint",
        action="store_true",
        help="save the estimator fit to file periodically while building. Only used if "
        "the estimator can checkpoint (default: %(default)s).",
    )
    parser.add_argument(
        "-rn",
        "--row_normalise",
        action="store_true",
        help="normalise the data rows prior to fitting and predicting. "
        "(default: %(default)s).",
    )
    parser.add_argument(
        "-nc",
        "--n_clusters",
        type=int,
        default=-1,
        help="the number of clusters to find for clusterers which have an {n_clusters} "
        "parameter. If {-1}, use the number of classes in the dataset "
        "(default: %(default)s).",
    )
    parser.add_argument(
        "-ctts",
        "--combine_test_train_split",
        action="store_true",
        help="whether to use a train/test split or not. If True, the train and test "
        "sets are combined and used the fit the estimator. Only available for "
        "clustering (default: %(default)s).",
    )
    parser.add_argument(
        "-bt",
        "--benchmark_time",
        action="store_true",
        help="run a benchmark function and save the time spent in the results file "
        "(default: %(default)s).",
    )
    parser.add_argument(
        "-wa",
        "--write_attributes",
        action="store_true",
        help="write the estimator attributes to file when running experiments. Will "
        "recursively write the attributes of sub-estimators if present. "
        "(default: %(default)s).",
    )
    parser.add_argument(
        "-ams",
        "--att_max_shape",
        type=int,
        default=0,
        help="The max estimator collections shape allowed when writing attributes, at "
        "0 no estimators in collections will be written, at 1 estimators in "
        "one-dimensional lists will be written etc. (default: %(default)s).",
    )
    parser.add_argument(
        "-kw",
        "--kwargs",
        "--kwarg",
        action="append",
        nargs=3,
        metavar=("KEY", "VALUE", "TYPE"),
        help="additional keyword arguments to pass to the estimator. Should contain "
        "the parameter to set, the parameter value, and the type of the value i.e. "
        "{--kwargs n_estimators 200 int} to change the size of an ensemble. Valid "
        "types are {int, float, bool, str}. Any other type will be passed as a str. "
        "Can be used multiple times (default: %(default)s).",
    )
    args = parser.parse_args(args=args)

    kwargs = {}
    if args.kwargs is not None:
        for kwarg in args.kwargs:
            if kwarg[2] == "int":
                kwargs[kwarg[0]] = int(kwarg[1])
            elif kwarg[2] == "float":
                kwargs[kwarg[0]] = float(kwarg[1])
            elif kwarg[2] == "bool":
                kwargs[kwarg[0]] = kwarg[1].lower() == "true" or kwarg[1] == "1"
            else:
                kwargs[kwarg[0]] = kwarg[1]
    args.kwargs = kwargs

    return args



"""Dataset sizes for testing."""

DATA_TRAIN_SIZES = {
    "EqualMinimalJapaneseVowels": 20,
    "MinimalCardanoSentiment": 20,
    "MinimalChinatown": 20,
    "MinimalGasPrices": 20,
    "MinimalJapaneseVowels": 20,
    "ShampooSales": 24,
    "UnequalMinimalCardanoSentiment": 20,
    "UnequalMinimalChinatown": 20,
    "UnequalMinimalGasPrices": 20,
}

DATA_TEST_SIZES = {
    "EqualMinimalJapaneseVowels": 20,
    "MinimalCardanoSentiment": 20,
    "MinimalChinatown": 20,
    "MinimalGasPrices": 20,
    "MinimalJapaneseVowels": 20,
    "ShampooSales": 12,
    "UnequalMinimalCardanoSentiment": 20,
    "UnequalMinimalChinatown": 20,
    "UnequalMinimalGasPrices": 20,
}



"""Unit test utilities."""

import os
import sys
from contextlib import contextmanager
from os import devnull
from pathlib import Path

from sklearn.base import BaseEstimator

_TEST_DATA_PATH = f"{os.path.dirname(Path(__file__).parent.parent)}/tsml_eval/datasets/"

_TEST_RESULTS_PATH = (
    f"{os.path.dirname(Path(__file__).parent.parent)}/tsml_eval/testing/"
    f"_test_results_files/"
)

_TEST_EVAL_PATH = (
    f"{os.path.dirname(Path(__file__).parent.parent)}/tsml_eval/testing/"
    f"_test_eval_files/"
)

_TEST_OUTPUT_PATH = f"{os.path.dirname(Path(__file__).parent.parent)}/test_output/"


def _check_set_method(
    set_method, estimator_sub_list, estimator_dict, all_estimator_names
):
    for estimator_names in estimator_sub_list:
        estimator_names = (
            [estimator_names] if isinstance(estimator_names, str) else estimator_names
        )

        for estimator_alias in estimator_names:
            assert (
                estimator_alias not in all_estimator_names
            ), f"Estimator {estimator_alias} is duplicated"
            all_estimator_names.append(estimator_alias)

            try:
                e = set_method(estimator_alias)
            except ModuleNotFoundError as err:
                exempt_errors = [
                    "optional dependency",
                    "soft dependency",
                    "python version",
                ]
                if any(s in str(err) for s in exempt_errors):
                    continue
                else:
                    raise err

            assert e is not None, f"Estimator {estimator_alias} not found"
            assert isinstance(
                e, BaseEstimator
            ), f"Estimator {estimator_alias} is not a BaseEstimator"

            c_name = e.__class__.__name__.lower()
            if c_name == estimator_alias.lower():
                estimator_dict[c_name] = True
            elif c_name not in estimator_dict:
                estimator_dict[c_name] = False


EXEMPT_ESTIMATOR_NAMES = [
    "channelensembleregressor",
    "gridsearchcv",
    "transformedtargetforecaster",
]


def _check_set_method_results(
    estimator_dict, estimator_name="Estimators", method_name="the method"
):
    for estimator in EXEMPT_ESTIMATOR_NAMES:
        if estimator in estimator_dict:
            estimator_dict.pop(estimator)

    if not all(estimator_dict.values()):
        missing_keys = [key for key, value in estimator_dict.items() if not value]
        raise ValueError(
            f"All {estimator_name.lower()} seen in {method_name} must have an entry "
            "for the full class name (usually with default parameters). "
            f"{estimator_name} with missing entries: {missing_keys}."
        )


@contextmanager
def suppress_output(suppress_stdout=True, suppress_stderr=True):
    """Redirects stdout and/or stderr to devnull."""
    with open(devnull, "w") as null:
        stdout = sys.stdout
        stderr = sys.stderr
        try:
            if suppress_stdout:
                sys.stdout = null
            if suppress_stderr:
                sys.stderr = null
            yield
        finally:
            if suppress_stdout:
                sys.stdout = stdout
            if suppress_stderr:
                sys.stderr = stderr



"""IVC consensus clustering algorithm from results files."""

import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.utils import check_random_state

from tsml_eval.estimators.clustering.consensus.ivc import IterativeVotingClustering


class FromFileIterativeVotingClustering(IterativeVotingClustering):
    """
    IVC (Iterative Voting Clustering) Consensus Clusterer.

    IVC is a consensus clustering algorithm that combines the results of multiple
    base clusterers to find a consensus clustering. It iteratively refines cluster
    assignments based on a majority voting scheme.

    Parameters
    ----------
    clusterers : list of str
        A list of paths to the clusterer result files to use in the ensemble.
    init : {'plus', 'random', 'aligned'}, default='plus'
        The method used to initialize the cluster centers. 'plus' uses the
        k-means++ initialization method, 'random' uses random selection, and
        'aligned' uses a method that aligns the cluster assignmeds from the base
        clusterers.
    n_clusters : int, default=8
        The number of clusters to form.
    max_iterations : int, default=500
        The maximum number of iterations to perform.
    overwrite_y : bool, default=False
        If True, the labels in the loaded files will overwrite the labels
        passed in the fit method.
    skip_y_check : bool, default=False
        If True, the labels in the loaded files will not be checked against
        the labels passed in the fit method.
    random_state : int, default=None
        The seed for random number generation.

    Attributes
    ----------
    labels_ : ndarray of shape (n_instances,)
        Labels of each point from the last fit.
    """

    def __init__(
        self,
        clusterers,
        init="plus",
        n_clusters=8,
        max_iterations=500,
        overwrite_y=False,
        skip_y_check=False,
        random_state=None,
    ):
        self.overwrite_y = overwrite_y
        self.skip_y_check = skip_y_check

        super().__init__(
            clusterers=clusterers,
            init=init,
            n_clusters=n_clusters,
            max_iterations=max_iterations,
            random_state=random_state,
        )

    def fit(self, X, y=None):
        """Fit model to X using IVC."""
        if isinstance(X, np.ndarray) and len(X.shape) == 3 and X.shape[1] == 1:
            X = np.reshape(X, (X.shape[0], -1))
        elif isinstance(X, pd.DataFrame) and len(X.shape) == 2:
            X = X.to_numpy()
        elif isinstance(X, list):
            X = np.array(X)
        elif not isinstance(X, np.ndarray) or len(X.shape) > 2:
            raise ValueError(
                "FromFileIterativeVotingClustering is not a time series clusterer. "
                "A valid sklearn input such as a 2d numpy array is required."
                "Sparse input formats are currently not supported."
            )
        X = self._validate_data(X=X, ensure_min_samples=self.n_clusters)

        # load train file at path (trainResample.csv if random_state is None,
        # trainResample{self.random_state}.csv otherwise)
        if self.random_state is not None:
            file_name = f"trainResample{self.random_state}.csv"
        else:
            file_name = "trainResample.csv"

        cluster_assignments = np.zeros(
            (len(self.clusterers), X.shape[0]), dtype=np.int32
        )
        for i, path in enumerate(self.clusterers):
            f = open(path + file_name)
            lines = f.readlines()
            line2 = lines[2].split(",")

            # verify file matches data
            if len(lines) - 3 != X.shape[0]:
                raise ValueError(
                    f"n_instances of {path + file_name} does not match X, "
                    f"expected {X.shape[0]}, got {len(lines) - 3}"
                )
            if (
                y is not None
                and not self.skip_y_check
                and len(np.unique(y)) != int(line2[5])
            ):
                raise ValueError(
                    f"n_classes of {path + file_name} does not match X, "
                    f"expected {len(np.unique(y))}, got {line2[6]}"
                )

            for j in range(X.shape[0]):
                line = lines[j + 3].split(",")

                if self.overwrite_y:
                    if i == 0:
                        y[j] = float(line[0])
                    elif not self.skip_y_check:
                        assert y[j] == float(line[0])
                elif y is not None and not self.skip_y_check:
                    if i == 0:
                        le = preprocessing.LabelEncoder()
                        y = le.fit_transform(y)
                    assert float(line[0]) == y[j]

                cluster_assignments[i][j] = int(line[1])

            uc = np.unique(cluster_assignments[i])
            if (np.sort(uc) != np.arange(self.n_clusters)).any():
                raise ValueError(
                    "Input clusterers must have cluster labels in the range "
                    "0 to  n_clusters - 1."
                )

        self._build_ensemble(cluster_assignments)

        return self

    def predict(self, X):
        """Predict cluster labels for X."""
        if isinstance(X, np.ndarray) and len(X.shape) == 3 and X.shape[1] == 1:
            X = np.reshape(X, (X.shape[0], -1))
        elif isinstance(X, pd.DataFrame) and len(X.shape) == 2:
            X = X.to_numpy()
        elif isinstance(X, list):
            X = np.array(X)
        elif not isinstance(X, np.ndarray) or len(X.shape) > 2:
            raise ValueError(
                "FromFileIterativeVotingClustering is not a time series clusterer. "
                "A valid sklearn input such as a 2d numpy array is required."
                "Sparse input formats are currently not supported."
            )
        X = self._validate_data(X=X, reset=False)

        # load train file at path (trainResample.csv if random_state is None,
        # trainResample{self.random_state}.csv otherwise)
        if self.random_state is not None:
            file_name = f"testResample{self.random_state}.csv"
        else:
            file_name = "testResample.csv"

        cluster_assignments = np.zeros(
            (len(self.clusterers), X.shape[0]), dtype=np.int32
        )
        for i, path in enumerate(self.clusterers):
            f = open(path + file_name)
            lines = f.readlines()

            # verify file matches data
            if len(lines) - 3 != len(X):
                if len(lines) - 3 != X.shape[0]:
                    raise ValueError(
                        f"n_instances of {path + file_name} does not match X, "
                        f"expected {X.shape[0]}, got {len(lines) - 3}"
                    )

            for j in range(len(X)):
                line = lines[j + 3].split(",")
                cluster_assignments[i][j] = int(line[1])

        rng = check_random_state(self.random_state)
        labels = self._calculate_cluster_membership(cluster_assignments, rng)

        return labels



"""Estimators that are not present in tsml or aeon."""

__all__ = [
    "SklearnToTsmlClassifier",
    "SklearnToTsmlClusterer",
    "SklearnToTsmlRegressor",
]

from tsml_eval.estimators.classification._sklearn_classifier import (
    SklearnToTsmlClassifier,
)
from tsml_eval.estimators.clustering._sklearn_clusterer import SklearnToTsmlClusterer
from tsml_eval.estimators.regression._sklearn_regressor import SklearnToTsmlRegressor



"""Functions to perform machine learning/data mining experiments.

Results are saved a standardised format used by tsml.
"""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]
__all__ = [
    "run_classification_experiment",
    "load_and_run_classification_experiment",
    "run_regression_experiment",
    "load_and_run_regression_experiment",
    "run_classification_experiment",
    "load_and_run_clustering_experiment",
    "run_forecasting_experiment",
    "load_and_run_forecasting_experiment",
]

import os
import time
import warnings
from datetime import datetime

import numpy as np
import pandas as pd
from aeon.classification import BaseClassifier
from aeon.clustering import BaseClusterer
from aeon.forecasting.base import BaseForecaster
from aeon.performance_metrics.clustering import clustering_accuracy_score
from aeon.regression.base import BaseRegressor
from aeon.transformations.collection import TimeSeriesScaler
from sklearn import preprocessing
from sklearn.base import BaseEstimator, is_classifier, is_regressor
from sklearn.metrics import (
    accuracy_score,
    mean_absolute_percentage_error,
    mean_squared_error,
)
from sklearn.model_selection import cross_val_predict
from tsml.base import BaseTimeSeriesEstimator
from tsml.utils.validation import is_clusterer

from tsml_eval.estimators import (
    SklearnToTsmlClassifier,
    SklearnToTsmlClusterer,
    SklearnToTsmlRegressor,
)
from tsml_eval.utils.datasets import load_experiment_data
from tsml_eval.utils.experiments import (
    _check_existing_results,
    estimator_attributes_to_file,
    timing_benchmark,
)
from tsml_eval.utils.memory_recorder import record_max_memory
from tsml_eval.utils.resampling import resample_data, stratified_resample_data
from tsml_eval.utils.results_writing import (
    write_classification_results,
    write_clustering_results,
    write_forecasting_results,
    write_regression_results,
)

if os.getenv("MEMRECORD_INTERVAL") is not None:  # pragma: no cover
    TEMP = os.getenv("MEMRECORD_INTERVAL")
    MEMRECORD_INTERVAL = float(TEMP) if isinstance(TEMP, str) else 5.0
else:
    MEMRECORD_INTERVAL = 5.0


def run_classification_experiment(
    X_train,
    y_train,
    X_test,
    y_test,
    classifier,
    results_path,
    row_normalise=False,
    classifier_name=None,
    dataset_name="N/A",
    resample_id=None,
    build_test_file=True,
    build_train_file=False,
    ignore_custom_train_estimate=False,
    attribute_file_path=None,
    att_max_shape=0,
    benchmark_time=True,
):
    """Run a classification experiment and save the results to file.

    Function to run a basic classification experiment for a
    <dataset>/<classifier>/<resample> combination and write the results to csv file(s)
    at a given location.

    Parameters
    ----------
    X_train : pd.DataFrame or np.array    todo
        The data to train the classifier.
    y_train : np.array
        Training data class labels.
    X_test : pd.DataFrame or np.array
        The data used to test the trained classifier.
    y_test : np.array
        Testing data class labels.
    classifier : BaseClassifier
        Classifier to be used in the experiment.
    results_path : str
        Location of where to write results. Any required directories will be created.
    row_normalise : bool, default=False
        Whether to normalise the data rows (time series) prior to fitting and
        predicting.
    classifier_name : str or None, default=None
        Name of classifier used in writing results. If None, the name is taken from
        the classifier.
    dataset_name : str, default="N/A"
        Name of dataset.
    resample_id : int or None, default=None
        Seed for resampling. If set to 0, the default train/test split from file is
        used. Also used in output file name.
    build_test_file : bool, default=True:
        Whether to generate test files or not. If the classifier can generate its own
        train probabilities, the classifier will be built but no file will be output.
    build_train_file : bool, default=False
        Whether to generate train files or not. If true, it performs a 10-fold
        cross-validation on the train data and saves. If the classifier can produce its
        own estimates, those are used instead.
    ignore_custom_train_estimate : bool, default=False
        todo
    attribute_file_path : str or None, default=None
        todo (only test)
    att_max_shape : int, default=0
        todo
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    """
    if not build_test_file and not build_train_file:
        raise ValueError(
            "Both test_file and train_file are set to False. "
            "At least one must be written."
        )

    if classifier_name is None:
        classifier_name = type(classifier).__name__

    use_fit_predict = False
    if isinstance(classifier, BaseClassifier):
        if not ignore_custom_train_estimate and classifier.get_tag(
            "capability:train_estimate", False, False
        ):
            use_fit_predict = True
    elif isinstance(classifier, BaseTimeSeriesEstimator) and is_classifier(classifier):
        pass
    elif isinstance(classifier, BaseEstimator) and is_classifier(classifier):
        classifier = SklearnToTsmlClassifier(
            classifier=classifier,
            pad_unequal=True,
            concatenate_channels=True,
            clone_estimator=False,
            random_state=(
                classifier.random_state if hasattr(classifier, "random_state") else None
            ),
        )
    else:
        raise TypeError("classifier must be a tsml, aeon or sklearn classifier.")

    if row_normalise:
        scaler = TimeSeriesScaler()
        X_train = scaler.fit_transform(X_train)
        X_test = scaler.fit_transform(X_test)

    le = preprocessing.LabelEncoder()
    y_train = le.fit_transform(y_train)
    y_test = le.transform(y_test)

    encoder_dict = {label: i for i, label in enumerate(le.classes_)}
    n_classes = len(np.unique(y_train))

    needs_fit = True
    fit_time = -1
    mem_usage = -1
    benchmark = -1
    train_time = -1
    fit_and_train_time = -1

    if benchmark_time:
        benchmark = timing_benchmark(random_state=resample_id)

    first_comment = (
        "Generated by run_classification_experiment on "
        f"{datetime.now().strftime('%m/%d/%Y, %H:%M:%S')}. "
        f"Encoder dictionary: {str(encoder_dict)}"
    )

    second = str(classifier.get_params()).replace("\n", " ").replace("\r", " ")

    if build_train_file:
        cv_size = 10
        start = int(round(time.time() * 1000))
        if use_fit_predict:
            train_probs = classifier.fit_predict_proba(X_train, y_train)
            needs_fit = False
            fit_and_train_time = int(round(time.time() * 1000)) - start
        else:
            _, counts = np.unique(y_train, return_counts=True)
            min_class = max(2, np.min(counts))
            if min_class < cv_size:
                cv_size = min_class

            train_probs = cross_val_predict(
                classifier, X_train, y=y_train, cv=cv_size, method="predict_proba"
            )
            train_time = int(round(time.time() * 1000)) - start

        train_preds = np.unique(y_train)[np.argmax(train_probs, axis=1)]
        train_acc = accuracy_score(y_train, train_preds)

        write_classification_results(
            train_preds,
            train_probs,
            y_train,
            classifier_name,
            dataset_name,
            results_path,
            full_path=False,
            split="TRAIN",
            resample_id=resample_id,
            time_unit="MILLISECONDS",
            first_line_comment=first_comment,
            parameter_info=second,
            accuracy=train_acc,
            fit_time=fit_time,
            predict_time=-1,
            benchmark_time=benchmark,
            memory_usage=mem_usage,
            n_classes=n_classes,
            train_estimate_method="Custom" if use_fit_predict else f"{cv_size}F-CV",
            train_estimate_time=train_time,
            fit_and_estimate_time=fit_and_train_time,
        )

    if build_test_file:
        if needs_fit:
            mem_usage, fit_time = record_max_memory(
                classifier.fit,
                args=(X_train, y_train),
                interval=MEMRECORD_INTERVAL,
                return_func_time=True,
            )
            fit_time += int(round(getattr(classifier, "_fit_time_milli", 0)))

        if attribute_file_path is not None:
            estimator_attributes_to_file(
                classifier, attribute_file_path, max_list_shape=att_max_shape
            )

        start = int(round(time.time() * 1000))
        test_probs = classifier.predict_proba(X_test)
        test_time = (
            int(round(time.time() * 1000))
            - start
            + int(round(getattr(classifier, "_predict_time_milli", 0)))
        )

        test_preds = classifier.classes_[np.argmax(test_probs, axis=1)]
        test_acc = accuracy_score(y_test, test_preds)

        write_classification_results(
            test_preds,
            test_probs,
            y_test,
            classifier_name,
            dataset_name,
            results_path,
            full_path=False,
            split="TEST",
            resample_id=resample_id,
            time_unit="MILLISECONDS",
            first_line_comment=first_comment,
            parameter_info=second,
            accuracy=test_acc,
            fit_time=fit_time,
            predict_time=test_time,
            benchmark_time=benchmark,
            memory_usage=mem_usage,
            n_classes=n_classes,
            train_estimate_method="N/A",
            train_estimate_time=-1,
            fit_and_estimate_time=fit_and_train_time,
        )


def load_and_run_classification_experiment(
    problem_path,
    results_path,
    dataset,
    classifier,
    row_normalise=False,
    classifier_name=None,
    resample_id=0,
    build_train_file=False,
    write_attributes=False,
    att_max_shape=0,
    benchmark_time=True,
    overwrite=False,
    predefined_resample=False,
):
    """Load a dataset and run a classification experiment.

    Function to load a dataset, run a basic classification experiment for a
    <dataset>/<classifier>/<resample> combination, and write the results to csv file(s)
    at a given location.

    Parameters
    ----------
    problem_path : str
        Location of problem files, full path.
    results_path : str
        Location of where to write results. Any required directories will be created.
    dataset : str
        Name of problem. Files must be <problem_path>/<dataset>/<dataset>+"_TRAIN.ts",
        same for "_TEST.ts".
    classifier : BaseClassifier
        Classifier to be used in the experiment.
    row_normalise : bool, default=False
        Whether to normalise the data rows (time series) prior to fitting and
        predicting.
    classifier_name : str or None, default=None
        Name of classifier used in writing results. If None, the name is taken from
        the classifier.
    resample_id : int, default=0
        Seed for resampling. If set to 0, the default train/test split from file is
        used. Also used in output file name.
    build_train_file : bool, default=False
        Whether to generate train files or not. If true, it performs a 10-fold
        cross-validation on the train data and saves. If the classifier can produce its
        own estimates, those are used instead.
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    overwrite : bool, default=False
        If set to False, this will only build results if there is not a result file
        already present. If True, it will overwrite anything already there.
    predefined_resample : bool, default=False
        Read a predefined resample from file instead of performing a resample. If True
        the file format must include the resample_id at the end of the dataset name i.e.
        <problem_path>/<dataset>/<dataset>+<resample_id>+"_TRAIN.ts".
    """
    if classifier_name is None:
        classifier_name = type(classifier).__name__

    build_test_file, build_train_file = _check_existing_results(
        results_path,
        classifier_name,
        dataset,
        resample_id,
        overwrite,
        True,
        build_train_file,
    )

    if not build_test_file and not build_train_file:
        warnings.warn("All files exist and not overwriting, skipping.", stacklevel=1)
        return

    X_train, y_train, X_test, y_test, resample = load_experiment_data(
        problem_path, dataset, resample_id, predefined_resample
    )

    if resample:
        X_train, y_train, X_test, y_test = stratified_resample_data(
            X_train, y_train, X_test, y_test, random_state=resample_id
        )

    if write_attributes:
        attribute_file_path = f"{results_path}/{classifier_name}/Workspace/{dataset}/"
    else:
        attribute_file_path = None

    run_classification_experiment(
        X_train,
        y_train,
        X_test,
        y_test,
        classifier,
        results_path,
        row_normalise=row_normalise,
        classifier_name=classifier_name,
        dataset_name=dataset,
        resample_id=resample_id,
        build_test_file=build_test_file,
        build_train_file=build_train_file,
        attribute_file_path=attribute_file_path,
        att_max_shape=att_max_shape,
        benchmark_time=benchmark_time,
    )


def run_regression_experiment(
    X_train,
    y_train,
    X_test,
    y_test,
    regressor,
    results_path,
    row_normalise=False,
    regressor_name=None,
    dataset_name="",
    resample_id=None,
    build_test_file=True,
    build_train_file=False,
    ignore_custom_train_estimate=False,
    attribute_file_path=None,
    att_max_shape=0,
    benchmark_time=True,
):
    """Run a regression experiment and save the results to file.

    Function to run a basic regression experiment for a
    <dataset>/<regressor>/<resample> combination and write the results to csv file(s)
    at a given location.

    Parameters
    ----------
    X_train : pd.DataFrame or np.array
        The data to train the regressor.
    y_train : np.array
        Training data labels.
    X_test : pd.DataFrame or np.array
        The data used to test the trained regressor.
    y_test : np.array
        Testing data labels.
    regressor : BaseRegressor
        Regressor to be used in the experiment.
    results_path : str
        Location of where to write results. Any required directories will be created.
    row_normalise : bool, default=False
        Whether to normalise the data rows (time series) prior to fitting and
        predicting.
    regressor_name : str or None, default=None
        Name of regressor used in writing results. If None, the name is taken from
        the regressor.
    dataset_name : str, default="N/A"
        Name of dataset.
    resample_id : int or None, default=None
        Seed for resampling. If set to 0, the default train/test split from file is
        used. Also used in output file name.
    build_test_file : bool, default=True:
        Whether to generate test files or not. If the regressor can generate its own
        train predictions, the classifier will be built but no file will be output.
    build_train_file : bool, default=False
        Whether to generate train files or not. If true, it performs a 10-fold
        cross-validation on the train data and saves. If the regressor can produce its
        own estimates, those are used instead.
    ignore_custom_train_estimate : bool, default=False
        todo
    attribute_file_path : str or None, default=None
        todo (only test)
    att_max_shape : int, default=0
        todo
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    """
    if not build_test_file and not build_train_file:
        raise ValueError(
            "Both test_file and train_file are set to False. "
            "At least one must be written."
        )

    if regressor_name is None:
        regressor_name = type(regressor).__name__

    use_fit_predict = False
    if isinstance(regressor, BaseRegressor):
        if not ignore_custom_train_estimate and regressor.get_tag(
            "capability:train_estimate", False, False
        ):
            use_fit_predict = True
    elif isinstance(regressor, BaseTimeSeriesEstimator) and is_regressor(regressor):
        pass
    elif isinstance(regressor, BaseEstimator) and is_regressor(regressor):
        regressor = SklearnToTsmlRegressor(
            regressor=regressor,
            pad_unequal=True,
            concatenate_channels=True,
            clone_estimator=False,
            random_state=(
                regressor.random_state if hasattr(regressor, "random_state") else None
            ),
        )
    else:
        raise TypeError("regressor must be a tsml, aeon or sklearn regressor.")

    if row_normalise:
        scaler = TimeSeriesScaler()
        X_train = scaler.fit_transform(X_train)
        X_test = scaler.fit_transform(X_test)

    needs_fit = True
    fit_time = -1
    mem_usage = -1
    benchmark = -1
    train_time = -1
    fit_and_train_time = -1

    if benchmark_time:
        benchmark = timing_benchmark(random_state=resample_id)

    first_comment = (
        "Generated by run_regression_experiment on "
        f"{datetime.now().strftime('%m/%d/%Y, %H:%M:%S')}"
    )

    second = str(regressor.get_params()).replace("\n", " ").replace("\r", " ")

    if build_train_file:
        cv_size = min(10, len(y_train))
        start = int(round(time.time() * 1000))
        if use_fit_predict:
            train_preds = regressor.fit_predict(X_train, y_train)
            needs_fit = False
            fit_and_train_time = int(round(time.time() * 1000)) - start
        else:
            train_preds = cross_val_predict(regressor, X_train, y=y_train, cv=cv_size)
            train_time = int(round(time.time() * 1000)) - start

        train_mse = mean_squared_error(y_train, train_preds)

        write_regression_results(
            train_preds,
            y_train,
            regressor_name,
            dataset_name,
            results_path,
            full_path=False,
            split="TRAIN",
            resample_id=resample_id,
            time_unit="MILLISECONDS",
            first_line_comment=first_comment,
            parameter_info=second,
            mse=train_mse,
            fit_time=fit_time,
            predict_time=-1,
            benchmark_time=benchmark,
            memory_usage=mem_usage,
            train_estimate_method="Custom" if use_fit_predict else f"{cv_size}F-CV",
            train_estimate_time=train_time,
            fit_and_estimate_time=fit_and_train_time,
        )

    if build_test_file:
        if needs_fit:
            mem_usage, fit_time = record_max_memory(
                regressor.fit,
                args=(X_train, y_train),
                interval=MEMRECORD_INTERVAL,
                return_func_time=True,
            )
            fit_time += int(round(getattr(regressor, "_fit_time_milli", 0)))

        if attribute_file_path is not None:
            estimator_attributes_to_file(
                regressor, attribute_file_path, max_list_shape=att_max_shape
            )

        start = int(round(time.time() * 1000))
        test_preds = regressor.predict(X_test)
        test_time = (int(round(time.time() * 1000)) - start) + int(
            round(getattr(regressor, "_predict_time_milli", 0))
        )

        test_mse = mean_squared_error(y_test, test_preds)

        write_regression_results(
            test_preds,
            y_test,
            regressor_name,
            dataset_name,
            results_path,
            full_path=False,
            split="TEST",
            resample_id=resample_id,
            time_unit="MILLISECONDS",
            first_line_comment=first_comment,
            parameter_info=second,
            mse=test_mse,
            fit_time=fit_time,
            predict_time=test_time,
            benchmark_time=benchmark,
            memory_usage=mem_usage,
            train_estimate_method="N/A",
            train_estimate_time=-1,
            fit_and_estimate_time=fit_and_train_time,
        )


def load_and_run_regression_experiment(
    problem_path,
    results_path,
    dataset,
    regressor,
    row_normalise=False,
    regressor_name=None,
    resample_id=0,
    build_train_file=False,
    write_attributes=False,
    att_max_shape=0,
    benchmark_time=True,
    overwrite=False,
    predefined_resample=False,
):
    """Load a dataset and run a regression experiment.

    Function to load a dataset, run a basic regression experiment for a
    <dataset>/<regressor>/<resample> combination, and write the results to csv file(s)
    at a given location.

    Parameters
    ----------
    problem_path : str
        Location of problem files, full path.
    results_path : str
        Location of where to write results. Any required directories will be created.
    dataset : str
        Name of problem. Files must be <problem_path>/<dataset>/<dataset>+"_TRAIN.ts",
        same for "_TEST.ts".
    regressor : BaseRegressor
        Regressor to be used in the experiment.
    row_normalise : bool, default=False
        Whether to normalise the data rows (time series) prior to fitting and
        predicting.
    regressor_name : str or None, default=None
        Name of regressor used in writing results. If None, the name is taken from
        the regressor.
    resample_id : int, default=0
        Seed for resampling. If set to 0, the default train/test split from file is
        used. Also used in output file name.
    build_train_file : bool, default=False
        Whether to generate train files or not. If true, it performs a 10-fold
        cross-validation on the train data and saves. If the regressor can produce its
        own estimates, those are used instead.
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    overwrite : bool, default=False
        If set to False, this will only build results if there is not a result file
        already present. If True, it will overwrite anything already there.
    predefined_resample : bool, default=False
        Read a predefined resample from file instead of performing a resample. If True
        the file format must include the resample_id at the end of the dataset name i.e.
        <problem_path>/<dataset>/<dataset>+<resample_id>+"_TRAIN.ts".
    """
    if regressor_name is None:
        regressor_name = type(regressor).__name__

    build_test_file, build_train_file = _check_existing_results(
        results_path,
        regressor_name,
        dataset,
        resample_id,
        overwrite,
        True,
        build_train_file,
    )

    if not build_test_file and not build_train_file:
        warnings.warn("All files exist and not overwriting, skipping.", stacklevel=1)
        return

    X_train, y_train, X_test, y_test, resample = load_experiment_data(
        problem_path, dataset, resample_id, predefined_resample
    )

    if resample:
        X_train, y_train, X_test, y_test = resample_data(
            X_train, y_train, X_test, y_test, random_state=resample_id
        )

    if write_attributes:
        attribute_file_path = f"{results_path}/{regressor_name}/Workspace/{dataset}/"
    else:
        attribute_file_path = None

    # Ensure labels are floats
    y_train = y_train.astype(float)
    y_test = y_test.astype(float)

    run_regression_experiment(
        X_train,
        y_train,
        X_test,
        y_test,
        regressor,
        results_path,
        row_normalise=row_normalise,
        regressor_name=regressor_name,
        dataset_name=dataset,
        resample_id=resample_id,
        build_test_file=build_test_file,
        build_train_file=build_train_file,
        attribute_file_path=attribute_file_path,
        att_max_shape=att_max_shape,
        benchmark_time=benchmark_time,
    )


def run_clustering_experiment(
    X_train,
    y_train,
    clusterer,
    results_path,
    X_test=None,
    y_test=None,
    row_normalise=False,
    n_clusters=None,
    clusterer_name=None,
    dataset_name="N/A",
    resample_id=None,
    build_test_file=False,
    build_train_file=True,
    attribute_file_path=None,
    att_max_shape=0,
    benchmark_time=True,
):
    """Run a clustering experiment and save the results to file.

    Function to run a basic clustering experiment for a
    <dataset>/<clusterer>/<resample> combination and write the results to csv file(s)
    at a given location.

    Parameters
    ----------
    X_train : pd.DataFrame or np.array
        The data to train the clusterer.
    y_train : np.array
        Training data class labels (used for evaluation).
    clusterer : BaseClusterer
        Clusterer to be used in the experiment.
    results_path : str
        Location of where to write results. Any required directories will be created.
    X_test : pd.DataFrame or np.array, default=None
        The data used to test the fitted clusterer.
    y_test : np.array, default=None
        Testing data class labels.
    row_normalise : bool, default=False
        Whether to normalise the data rows (time series) prior to fitting and
        predicting.
    n_clusters : int or None, default=None
        Number of clusters to use if the clusterer has an `n_clusters` parameter.
        If None, the clusterers default is used. If -1, the number of classes in the
        dataset is used.

        This may not work as intended for pipelines currently.
    clusterer_name : str or None, default=None
        Name of clusterer used in writing results. If None, the name is taken from
        the clusterer.
    dataset_name : str, default="N/A"
        Name of dataset.
    resample_id : int or None, default=None
        Seed for resampling. If set to 0, the default train/test split from file is
        used. Also used in output file name.
    build_test_file : bool, default=False:
        Whether to generate test files or not. If True, X_test and y_test must be
        provided.
    build_train_file : bool, default=True
        Whether to generate train files or not. The clusterer is fit using train data
        regardless of input.
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    """
    if not build_test_file and not build_train_file:
        raise ValueError(
            "Both test_file and train_file are set to False. "
            "At least one must be written."
        )

    if clusterer_name is None:
        clusterer_name = type(clusterer).__name__

    if isinstance(clusterer, BaseClusterer) or (
        isinstance(clusterer, BaseTimeSeriesEstimator) and is_clusterer(clusterer)
    ):
        pass
    elif isinstance(clusterer, BaseEstimator) and is_clusterer(clusterer):
        clusterer = SklearnToTsmlClusterer(
            clusterer=clusterer,
            pad_unequal=True,
            concatenate_channels=True,
            clone_estimator=False,
            random_state=(
                clusterer.random_state if hasattr(clusterer, "random_state") else None
            ),
        )
    else:
        raise TypeError("clusterer must be a tsml, aeon or sklearn clusterer.")

    if build_test_file and (X_test is None or y_test is None):
        raise ValueError("Test data and labels not provided, cannot build test file.")

    if row_normalise:
        scaler = TimeSeriesScaler()
        X_train = scaler.fit_transform(X_train)
        if build_test_file:
            X_test = scaler.fit_transform(X_test)

    le = preprocessing.LabelEncoder()
    y_train = le.fit_transform(y_train)
    if build_test_file:
        y_test = le.transform(y_test)

    encoder_dict = {label: i for i, label in enumerate(le.classes_)}
    n_classes = len(np.unique(y_train))

    benchmark = -1
    if benchmark_time:
        benchmark = timing_benchmark(random_state=resample_id)

    first_comment = (
        "Generated by run_clustering_experiment on "
        f"{datetime.now().strftime('%m/%d/%Y, %H:%M:%S')}. "
        f"Encoder dictionary: {str(encoder_dict)}"
    )

    if isinstance(n_clusters, int):
        try:
            if n_clusters == -1:
                n_clusters = n_classes

            if isinstance(clusterer, SklearnToTsmlClusterer):
                clusterer.set_params(clusterer__n_clusters=n_clusters)
            else:
                clusterer.set_params(n_clusters=n_clusters)
        except ValueError:
            warnings.warn(
                f"{clusterer_name} does not have a n_clusters parameter, "
                "so it cannot be set.",
                stacklevel=1,
            )
            n_clusters = None
    elif n_clusters is not None:
        raise ValueError("n_clusters must be an int or None.")

    second = str(clusterer.get_params()).replace("\n", " ").replace("\r", " ")

    mem_usage, fit_time = record_max_memory(
        clusterer.fit,
        args=(X_train,),
        interval=MEMRECORD_INTERVAL,
        return_func_time=True,
    )
    fit_time += int(round(getattr(clusterer, "_fit_time_milli", 0)))

    if attribute_file_path is not None:
        estimator_attributes_to_file(
            clusterer, attribute_file_path, max_list_shape=att_max_shape
        )

    start = int(round(time.time() * 1000))
    if callable(getattr(clusterer, "predict_proba", None)):
        train_probs = clusterer.predict_proba(X_train)
        train_preds = np.argmax(train_probs, axis=1)
    else:
        train_preds = (
            clusterer.labels_
            if hasattr(clusterer, "labels_")
            else clusterer.predict(X_train)
        )
        train_probs = np.zeros(
            (
                len(train_preds),
                n_clusters if n_clusters is not None else len(np.unique(train_preds)),
            )
        )
        train_probs[np.arange(len(train_preds)), train_preds] = 1
    train_time = int(round(time.time() * 1000)) - start

    if build_train_file:
        train_acc = clustering_accuracy_score(y_train, train_preds)

        write_clustering_results(
            train_preds,
            train_probs,
            y_train,
            clusterer_name,
            dataset_name,
            results_path,
            full_path=False,
            split="TRAIN",
            resample_id=resample_id,
            time_unit="MILLISECONDS",
            first_line_comment=first_comment,
            parameter_info=second,
            clustering_accuracy=train_acc,
            fit_time=fit_time,
            predict_time=train_time,
            benchmark_time=benchmark,
            memory_usage=mem_usage,
            n_classes=n_classes,
            n_clusters=len(train_probs[0]),
        )

    if build_test_file:
        start = int(round(time.time() * 1000))
        if callable(getattr(clusterer, "predict_proba", None)):
            test_probs = clusterer.predict_proba(X_test)
            test_preds = np.argmax(test_probs, axis=1)
        else:
            test_preds = clusterer.predict(X_test)
            test_probs = np.zeros(
                (
                    len(test_preds),
                    (
                        n_clusters
                        if n_clusters is not None
                        else len(np.unique(train_preds))
                    ),
                )
            )
            test_probs[np.arange(len(test_preds)), test_preds] = 1
        test_time = (
            int(round(time.time() * 1000))
            - start
            + int(round(getattr(clusterer, "_predict_time_milli", 0)))
        )

        test_acc = clustering_accuracy_score(y_test, test_preds)

        write_clustering_results(
            test_preds,
            test_probs,
            y_test,
            clusterer_name,
            dataset_name,
            results_path,
            full_path=False,
            split="TEST",
            resample_id=resample_id,
            time_unit="MILLISECONDS",
            first_line_comment=first_comment,
            parameter_info=second,
            clustering_accuracy=test_acc,
            fit_time=fit_time,
            predict_time=test_time,
            benchmark_time=benchmark,
            memory_usage=mem_usage,
            n_classes=n_classes,
            n_clusters=len(test_probs[0]),
        )


def load_and_run_clustering_experiment(
    problem_path,
    results_path,
    dataset,
    clusterer,
    row_normalise=False,
    n_clusters=None,
    clusterer_name=None,
    resample_id=0,
    build_test_file=False,
    write_attributes=False,
    att_max_shape=0,
    benchmark_time=True,
    overwrite=False,
    predefined_resample=False,
    combine_train_test_split=False,
):
    """Load a dataset and run a clustering experiment.

    Function to load a dataset, run a basic clustering experiment for a
    <dataset>/<clusterer>/<resample> combination, and write the results to csv file(s)
    at a given location.

    Parameters
    ----------
    problem_path : str
        Location of problem files, full path.
    results_path : str
        Location of where to write results. Any required directories will be created.
    dataset : str
        Name of problem. Files must be <problem_path>/<dataset>/<dataset>+"_TRAIN.ts",
        same for "_TEST.ts".
    clusterer : BaseClusterer
        Clusterer to be used in the experiment.
    row_normalise : bool, default=False
        Whether to normalise the data rows (time series) prior to fitting and
        predicting.
    n_clusters : int or None, default=None
        Number of clusters to use if the clusterer has an `n_clusters` parameter.
        If None, the clusterers default is used. If -1, the number of classes in the
        dataset is used.
    clusterer_name : str or None, default=None
        Name of clusterer used in writing results. If None, the name is taken from
        the clusterer.
    resample_id : int, default=0
        Seed for resampling. If set to 0, the default train/test split from file is
        used. Also used in output file name.
    build_test_file : bool, default=False
        Whether to generate test files or not. If true, the clusterer will assign
        clusters to the loaded test data.
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    overwrite : bool, default=False
        If set to False, this will only build results if there is not a result file
        already present. If True, it will overwrite anything already there.
    predefined_resample : bool, default=False
        Read a predefined resample from file instead of performing a resample. If True
        the file format must include the resample_id at the end of the dataset name i.e.
        <problem_path>/<dataset>/<dataset>+<resample_id>+"_TRAIN.ts".
    combine_train_test_split: bool, default=False
        Whether the train/test split should be combined. If True then
        the train/test split is combined into a single train set. If False then the
        train/test split is used as normal.
    """
    if clusterer_name is None:
        clusterer_name = type(clusterer).__name__

    if combine_train_test_split:
        build_test_file = False

    build_test_file, build_train_file = _check_existing_results(
        results_path,
        clusterer_name,
        dataset,
        resample_id,
        overwrite,
        build_test_file,
        True,
    )

    if not build_test_file and not build_train_file:
        warnings.warn("All files exist and not overwriting, skipping.", stacklevel=1)
        return

    X_train, y_train, X_test, y_test, resample = load_experiment_data(
        problem_path, dataset, resample_id, predefined_resample
    )

    if resample:
        X_train, y_train, X_test, y_test = stratified_resample_data(
            X_train, y_train, X_test, y_test, random_state=resample_id
        )

    if write_attributes:
        attribute_file_path = f"{results_path}/{clusterer_name}/Workspace/{dataset}/"
    else:
        attribute_file_path = None

    if combine_train_test_split:
        y_train = np.concatenate((y_train, y_test), axis=None)
        X_train = (
            np.concatenate([X_train, X_test], axis=0)
            if isinstance(X_train, np.ndarray)
            else X_train + X_test
        )
        X_test = None
        y_test = None

    run_clustering_experiment(
        X_train,
        y_train,
        clusterer,
        results_path,
        X_test=X_test,
        y_test=y_test,
        row_normalise=row_normalise,
        n_clusters=n_clusters,
        clusterer_name=clusterer_name,
        dataset_name=dataset,
        resample_id=resample_id,
        build_train_file=build_train_file,
        build_test_file=build_test_file,
        attribute_file_path=attribute_file_path,
        att_max_shape=att_max_shape,
        benchmark_time=benchmark_time,
    )


def run_forecasting_experiment(
    train,
    test,
    forecaster,
    results_path,
    forecaster_name=None,
    dataset_name="N/A",
    random_seed=None,
    attribute_file_path=None,
    att_max_shape=0,
    benchmark_time=True,
):
    """Run a forecasting experiment and save the results to file.

    Function to run a basic forecasting experiment for a
    <dataset>/<forecaster>/<resample> combination and write the results to csv file(s)
    at a given location.

    Parameters
    ----------
    train : pd.DataFrame or np.array
        The series used to train the forecaster.
    test : pd.DataFrame or np.array
        The series used to test the trained forecaster.
    forecaster : BaseForecaster
        Regressor to be used in the experiment.
    results_path : str
        Location of where to write results. Any required directories will be created.
    forecaster_name : str or None, default=None
        Name of forecaster used in writing results. If None, the name is taken from
        the forecaster.
    dataset_name : str, default="N/A"
        Name of dataset.
    random_seed : int or None, default=None
        Indicates what random seed was used as a random_state for the forecaster. Only
        used for the results file name.
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    """
    if not isinstance(forecaster, BaseForecaster):
        raise TypeError("forecaster must be an aeon forecaster.")

    if forecaster_name is None:
        forecaster_name = type(forecaster).__name__

    benchmark = -1
    if benchmark_time:
        benchmark = timing_benchmark(random_state=random_seed)

    first_comment = (
        "Generated by run_forecasting_experiment on "
        f"{datetime.now().strftime('%m/%d/%Y, %H:%M:%S')}"
    )

    second = str(forecaster.get_params()).replace("\n", " ").replace("\r", " ")

    mem_usage, fit_time = record_max_memory(
        forecaster.fit,
        args=(train,),
        interval=MEMRECORD_INTERVAL,
        return_func_time=True,
    )
    fit_time += int(round(getattr(forecaster, "_fit_time_milli", 0)))

    if attribute_file_path is not None:
        estimator_attributes_to_file(
            forecaster, attribute_file_path, max_list_shape=att_max_shape
        )

    start = int(round(time.time() * 1000))
    test_preds = forecaster.predict(np.arange(1, len(test) + 1))
    test_time = (
        int(round(time.time() * 1000))
        - start
        + int(round(getattr(forecaster, "_predict_time_milli", 0)))
    )
    test_preds = test_preds.flatten()

    test_mape = mean_absolute_percentage_error(test, test_preds)

    write_forecasting_results(
        test_preds,
        test,
        forecaster_name,
        dataset_name,
        results_path,
        full_path=False,
        split="TEST",
        random_seed=random_seed,
        time_unit="MILLISECONDS",
        first_line_comment=first_comment,
        parameter_info=second,
        mape=test_mape,
        fit_time=fit_time,
        predict_time=test_time,
        benchmark_time=benchmark,
        memory_usage=mem_usage,
    )


def load_and_run_forecasting_experiment(
    problem_path,
    results_path,
    dataset,
    forecaster,
    forecaster_name=None,
    random_seed=None,
    write_attributes=False,
    att_max_shape=0,
    benchmark_time=True,
    overwrite=False,
):
    """Load a dataset and run a regression experiment.

    Function to load a dataset, run a basic regression experiment for a
    <dataset>/<regressor/<resample> combination, and write the results to csv file(s)
    at a given location.

    Parameters
    ----------
    problem_path : str
        Location of problem files, full path.
    results_path : str
        Location of where to write results. Any required directories will be created.
    dataset : str
        Name of problem. Files must be <problem_path>/<dataset>/<dataset>+"_TRAIN.csv",
        same for "_TEST.csv".
    forecaster : BaseForecaster
        Regressor to be used in the experiment.
    forecaster_name : str or None, default=None
        Name of forecaster used in writing results. If None, the name is taken from
        the forecaster.
    random_seed : int or None, default=None
        Indicates what random seed was used as a random_state for the forecaster. Only
        used for the results file name.
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    overwrite : bool, default=False
        If set to False, this will only build results if there is not a result file
        already present. If True, it will overwrite anything already there.
    """
    if forecaster_name is None:
        forecaster_name = type(forecaster).__name__

    build_test_file, _ = _check_existing_results(
        results_path,
        forecaster_name,
        dataset,
        random_seed,
        overwrite,
        True,
        False,
    )

    if not build_test_file:
        warnings.warn("All files exist and not overwriting, skipping.", stacklevel=1)
        return

    if write_attributes:
        attribute_file_path = f"{results_path}/{forecaster_name}/Workspace/{dataset}/"
    else:
        attribute_file_path = None

    train = pd.read_csv(
        f"{problem_path}/{dataset}/{dataset}_TRAIN.csv", index_col=0
    ).squeeze("columns")
    train = train.astype(float).to_numpy()
    test = pd.read_csv(
        f"{problem_path}/{dataset}/{dataset}_TEST.csv", index_col=0
    ).squeeze("columns")
    test = test.astype(float).to_numpy()

    run_forecasting_experiment(
        train,
        test,
        forecaster,
        results_path,
        forecaster_name=forecaster_name,
        dataset_name=dataset,
        random_seed=random_seed,
        attribute_file_path=attribute_file_path,
        att_max_shape=att_max_shape,
        benchmark_time=benchmark_time,
    )



"""Regressors used in the publication."""

__author__ = ["MatthewMiddlehurst"]

import numpy as np

from tsml_eval.utils.functions import str_in_nested_list

rist_regressors = [
    ["inceptione", "inception-e", "inceptiontime", "InceptionTimeRegressor"],
    ["rocket", "RocketRegressor"],
    ["drcif", "DrCIFRegressor"],
    ["fresh-prince", "freshprince", "FreshPRINCERegressor"],
    ["RDSTRegressor", "rdst"],
    ["RISTRegressor", "rist", "rist-extrat"],
    "rist-rf",
    "rist-ridgecv",
]


def _set_rist_regressor(
    regressor_name,
    random_state=None,
    n_jobs=1,
    **kwargs,
):
    r = regressor_name.lower()

    if not str_in_nested_list(rist_regressors, r):
        raise ValueError(f"UNKNOWN REGRESSOR: {r} in _set_rist_regressor")

    if (
        r == "inceptione"
        or r == "inception-e"
        or r == "inceptiontime"
        or r == "inceptiontimeregressor"
    ):
        from aeon.regression.deep_learning import InceptionTimeRegressor

        return InceptionTimeRegressor(random_state=random_state, **kwargs)
    elif r == "rocket" or r == "rocketregressor":
        from aeon.regression.convolution_based import RocketRegressor

        return RocketRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif r == "drcif" or r == "drcifregressor":
        from aeon.regression.interval_based import DrCIFRegressor

        return DrCIFRegressor(
            n_estimators=500, random_state=random_state, n_jobs=n_jobs, **kwargs
        )
    elif r == "fresh-prince" or r == "freshprince" or r == "freshprinceregressor":
        from aeon.regression.feature_based import FreshPRINCERegressor

        return FreshPRINCERegressor(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "rdstregressor" or r == "rdst":
        from tsml.shapelet_based import RDSTRegressor

        return RDSTRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif r == "ristregressor" or r == "rist" or r == "rist-extrat":
        from sklearn.ensemble import ExtraTreesRegressor
        from tsml.hybrid import RISTRegressor

        return RISTRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=ExtraTreesRegressor(n_estimators=500),
            **kwargs,
        )
    elif r == "rist-rf":
        from sklearn.ensemble import RandomForestRegressor
        from tsml.hybrid import RISTRegressor

        return RISTRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=RandomForestRegressor(n_estimators=500),
            **kwargs,
        )
    elif r == "rist-ridgecv":
        from sklearn.linear_model import RidgeCV
        from sklearn.pipeline import make_pipeline
        from sklearn.preprocessing import StandardScaler
        from tsml.hybrid import RISTRegressor

        return RISTRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=make_pipeline(
                StandardScaler(with_mean=False),
                RidgeCV(alphas=np.logspace(-4, 4, 20)),
            ),
            **kwargs,
        )



"""Experiment runner for distance clustering publication."""

__author__ = ["MatthewMiddlehurst"]

import os
import sys

os.environ["MKL_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["NUMEXPR_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["OMP_NUM_THREADS"] = "1"  # must be done before numpy import!!

from tsml.base import _clone_estimator

from tsml_eval.experiments import load_and_run_clustering_experiment
from tsml_eval.publications.y2023.distance_based_clustering.set_distance_clusterer import (  # noqa: E501
    _set_distance_clusterer,
)
from tsml_eval.publications.y2023.distance_based_clustering.tests import (
    _DISTANCE_TEST_RESULTS_PATH,
)
from tsml_eval.testing.testing_utils import _TEST_DATA_PATH
from tsml_eval.utils.arguments import parse_args
from tsml_eval.utils.experiments import _results_present

classifiers = [
    "KMeans-dtw",
    "KMeans-ddtw",
    "KMeans-ed",
    "KMeans-edr",
    "KMeans-erp",
    "KMeans-lcss",
    "KMeans-msm",
    "KMeans-twe",
    "KMeans-wdtw",
    "KMeans-wddtw",
    "KMedoids-dtw",
    "KMedoids-ddtw",
    "KMedoids-ed",
    "KMedoids-edr",
    "KMedoids-erp",
    "KMedoids-lcss",
    "KMedoids-msm",
    "KMedoids-twe",
    "KMedoids-wdtw",
    "KMedoids-wddtw",
]


def _run_experiment(args):
    if args is None or args.__len__() < 1:
        data_path = _TEST_DATA_PATH
        results_path = _DISTANCE_TEST_RESULTS_PATH
        clusterer = "KMeans-dtw"
        dataset_name = "MinimalChinatown"
        resample_id = 0
        normalise = False
        kwargs = {}
        overwrite = False
    else:
        print("Input args = ", args)
        args = parse_args(args)
        data_path = args.data_path
        results_path = args.results_path
        clusterer = args.estimator_name
        dataset_name = args.dataset_name
        resample_id = args.resample_id
        normalise = args.row_normalise
        kwargs = args.kwargs
        overwrite = args.overwrite

    distance = clusterer.split("-")[-1]

    # further default parameterisation for clusterers and distances.
    # feel free to change
    kwargs["init_algorithm"] = "random"
    kwargs["max_iter"] = 30
    kwargs["n_init"] = 10

    distance_params = {
        "window": 0.2 if distance == "dtw" or distance == "wdtw" else 1.0,
        "epsilon": 0.05,
        "g": 0.05,
        "c": 1.0,
        "nu": 0.05,
        "lmbda": 1.0,
        "strategy": "independent",
    }
    kwargs["distance_params"] = distance_params

    cnl = clusterer.lower()
    if cnl.find("kmeans") or cnl.find("k-means"):
        kwargs["averaging_method"] = "mean"
        average_params = {
            **distance_params,
            "averaging_distance_metric": distance,
            "medoids_distance_metric": distance,
        }
        kwargs["average_params"] = average_params

    # Skip if not overwrite and results already present
    # this is also checked in load_and_run, but doing a quick check here so can
    # print a message and make sure data is not loaded
    if not overwrite and _results_present(
        results_path,
        clusterer,
        dataset_name,
        resample_id=resample_id,
        split="BOTH",
    ):
        print("Ignoring, results already present")
    else:
        load_and_run_clustering_experiment(
            data_path,
            results_path,
            dataset_name,
            (
                _set_distance_clusterer(
                    clusterer,
                    random_state=resample_id + 1,
                    **kwargs,
                )
                if isinstance(clusterer, str)
                else _clone_estimator(clusterer, resample_id)
            ),
            row_normalise=normalise,
            n_clusters=-1,
            clusterer_name=clusterer,
            resample_id=resample_id,
            build_test_file=True,
            overwrite=overwrite,
        )


if __name__ == "__main__":
    """
    Example simple usage, with arguments input via script or hard coded for testing.

    1. Edit the arguments from line 49-56 to suit your experiment, The most important
       are:
         data_path: the path to the data
         results_path: the path to the results
         clusterer: the name of the clusterer to use (check set_distance_clusterer.py),
         or an estimator object
         resample_id: the data resample id and random seed to use
    2. Run the script, if the experiment runs successfully a set of folders and a
       results csv file will be created in the results path.

    For evaluation of the written results, you can use the evaluation package, see
    our examples for usage:
    https://github.com/time-series-machine-learning/tsml-eval/blob/main/examples/

    For using your own clusterer, any clusterer following the sklearn, aeon,
    or tsml interface should be compatible with this file.
    """
    print("Running run_distance_experiments.py main")
    _run_experiment(sys.argv[1:])



"""Class for storing and loading results from a clustering experiment."""

import numpy as np
from aeon.performance_metrics.clustering import clustering_accuracy_score
from numpy.testing import assert_allclose
from sklearn.metrics import (
    adjusted_mutual_info_score,
    adjusted_rand_score,
    mutual_info_score,
    normalized_mutual_info_score,
    rand_score,
)

from tsml_eval.evaluation.storage.estimator_results import EstimatorResults
from tsml_eval.utils.results_writing import write_clustering_results


class ClustererResults(EstimatorResults):
    """
    A class for storing and managing results from clustering experiments.

    This class provides functionalities for storing clustering results,
    including cluster labels, probabilities, and various performance metrics.
    It extends the `EstimatorResults` class, inheriting its base functionalities.

    Parameters
    ----------
    dataset_name : str, default="N/A"
        Name of the dataset used.
    clusterer_name : str, default="N/A"
        Name of the clusterer used.
    split : str, default="N/A"
        Type of data split used, i.e. "train" or "test".
    resample_id : int or None, default=None
        Random seed used for the data resample, with 0 usually being the original data.
    time_unit : str, default="nanoseconds"
        Time measurement used for other fields.
    description : str, default=""
        Additional description of the clustering experiment. Appended to the end
        of the first line of the results file.
    parameters : str, default="No parameter info"
        Information about parameters used in the clusterer and other build information.
        Written to the second line of the results file.
    fit_time : float, default=-1.0
        Time taken fitting the model.
    predict_time : float, default=-1.0
        Time taken making predictions.
    benchmark_time : float, default=-1.0
        Time taken to run a simple benchmark function. In tsml-eval experiments, this
        is the time spent to sort 1,000 (seeded) random numpy arrays of size 20,000.
    memory_usage : float, default=-1.0
        Memory usage during the experiment. In tsml-eval experiments, this is the peak
        memory usage during the fit method.
    n_classes : int or None, default=None
        Number of classes in the dataset.
    n_clusters : int or None, default=None
        Number of clusters generated.
    class_labels : array-like or None, default=None
        Actual class labels.
    predictions : array-like or None, default=None
        Predicted cluster labels.
    probabilities : array-like or None, default=None
        Predicted cluster probabilities.
    pred_times : array-like or None, default=None
        Prediction times for each case.
    pred_descriptions : list of str or None, default=None
        Descriptions for each prediction.

    Attributes
    ----------
    n_cases : int or None
        Number of cases in the dataset.
    clustering_accuracy : float or None
        Clustering accuracy score.
    rand_index : float or None
        Rand score.
    adjusted_rand_index : float or None
        Adjusted Rand score.
    mutual_information : float or None
        Mutual information score.
    adjusted_mutual_information : float or None
        Adjusted mutual information score.
    normalised_mutual_information : float or None
        Normalised mutual information score.

    Examples
    --------
    >>> from tsml_eval.evaluation.storage import ClustererResults
    >>> from tsml_eval.testing.testing_utils import _TEST_RESULTS_PATH
    >>> cr = ClustererResults().load_from_file(
    ...     _TEST_RESULTS_PATH +
    ...     "/clustering/KMeans/Predictions/Trace/trainResample0.csv"
    ... )
    >>> cr.calculate_statistics()
    >>> cr.clustering_accuracy
    0.57
    """

    def __init__(
        self,
        dataset_name="N/A",
        clusterer_name="N/A",
        split="N/A",
        resample_id=None,
        time_unit="nanoseconds",
        description="",
        parameters="No parameter info",
        fit_time=-1.0,
        predict_time=-1.0,
        benchmark_time=-1.0,
        memory_usage=-1.0,
        n_classes=None,
        n_clusters=None,
        class_labels=None,
        predictions=None,
        probabilities=None,
        pred_times=None,
        pred_descriptions=None,
    ):
        # Line 1
        self.clusterer_name = clusterer_name

        # Line 3
        self.n_classes = n_classes
        self.n_clusters = n_clusters

        # Results
        self.class_labels = class_labels
        self.predictions = predictions
        self.probabilities = probabilities
        self.pred_times = pred_times
        self.pred_descriptions = pred_descriptions

        self.n_cases = None

        self.clustering_accuracy = None
        self.rand_index = None
        self.adjusted_rand_index = None
        self.mutual_information = None
        self.adjusted_mutual_information = None
        self.normalised_mutual_information = None

        super().__init__(
            dataset_name=dataset_name,
            estimator_name=clusterer_name,
            split=split,
            resample_id=resample_id,
            time_unit=time_unit,
            description=description,
            parameters=parameters,
            fit_time=fit_time,
            predict_time=predict_time,
            benchmark_time=benchmark_time,
            memory_usage=memory_usage,
        )

    # var_name: (display_name, higher is better, is timing)
    statistics = {
        "clustering_accuracy": ("CLAcc", True, False),
        "rand_index": ("RI", True, False),
        "adjusted_rand_index": ("ARI", True, False),
        "mutual_information": ("MI", True, False),
        "adjusted_mutual_information": ("AMI", True, False),
        "normalised_mutual_information": ("NMI", True, False),
        **EstimatorResults.statistics,
    }

    def save_to_file(self, file_path, full_path=True):
        """
        Write the clusterer results into a file format used by tsml.

        Parameters
        ----------
        file_path : str
            Path to write the results file to or the directory to build the default file
            structure if full_path is False.
        full_path : boolean, default=True
            If True, results are written directly to the directory passed in file_path.
            If False, then a standard file structure using the clusterer and dataset
            names is created and used to write the results file.
        """
        self.infer_size()

        if self.clustering_accuracy is None:
            self.clustering_accuracy = clustering_accuracy_score(
                self.class_labels, self.predictions
            )

        write_clustering_results(
            self.predictions,
            self.probabilities,
            self.class_labels,
            self.estimator_name,
            self.dataset_name,
            file_path,
            full_path=full_path,
            split=self.split,
            resample_id=self.resample_id,
            time_unit=self.time_unit,
            first_line_comment=self.description,
            parameter_info=self.parameter_info,
            clustering_accuracy=self.clustering_accuracy,
            fit_time=self.fit_time,
            predict_time=self.predict_time,
            benchmark_time=self.benchmark_time,
            memory_usage=self.memory_usage,
            n_classes=self.n_classes,
            n_clusters=self.n_clusters,
        )

    def load_from_file(self, file_path, verify_values=True):
        """
        Load clusterer results from a specified file.

        This method reads a file containing clusterer results and reconstructs the
        ClustererResults object. It calculates performance statistics and
        verifies values based on the loaded data.

        Parameters
        ----------
        file_path : str
            The path to the file from which clusterer results should be loaded. The
            file should be a tsml formatted clusterer results file.
        verify_values : bool, default=True
            If the method should perform verification of the loaded values.

        Returns
        -------
        self : ClustererResults
            The same ClustererResults object with loaded results.
        """
        cr = load_clusterer_results(file_path, verify_values=verify_values)
        self.__dict__.update(cr.__dict__)
        return self

    def calculate_statistics(self, overwrite=False):
        """
        Calculate various performance statistics based on the clusterer results.

        This method computes various performance metrics, such as clustering accuracy,
        Rand score, and others, based on the clusterers output.

        Parameters
        ----------
        overwrite : bool, default=False
            If the function should overwrite the current values when they are not None.
        """
        self.infer_size(overwrite=overwrite)

        if self.clustering_accuracy is None or overwrite:
            self.clustering_accuracy = clustering_accuracy_score(
                self.class_labels, self.predictions
            )
        if self.rand_index is None or overwrite:
            self.rand_index = rand_score(self.class_labels, self.predictions)
        if self.adjusted_rand_index is None or overwrite:
            self.adjusted_rand_index = adjusted_rand_score(
                self.class_labels, self.predictions
            )
        if self.mutual_information is None or overwrite:
            self.mutual_information = mutual_info_score(
                self.class_labels, self.predictions
            )
        if self.adjusted_mutual_information is None or overwrite:
            self.adjusted_mutual_information = adjusted_mutual_info_score(
                self.class_labels, self.predictions
            )
        if self.normalised_mutual_information is None or overwrite:
            self.normalised_mutual_information = normalized_mutual_info_score(
                self.class_labels, self.predictions
            )

    def infer_size(self, overwrite=False):
        """
        Infer and return the size of the dataset used in the results.

        This method estimates the size of the dataset that was used for the estimator,
        based on the results data.

        Also infers the number of clusters generated.

        Parameters
        ----------
        overwrite : bool, default=False
            If the function should overwrite the current values when they are not None.
        """
        if self.n_cases is None or overwrite:
            self.n_cases = len(self.class_labels)
        if self.n_clusters is None or overwrite:
            self.n_clusters = len(self.probabilities[0])


def load_clusterer_results(file_path, calculate_stats=True, verify_values=True):
    """
    Load and return clusterer results from a specified file.

    This function reads a file containing clusterer results and reconstructs the
    ClustererResults object. It optionally calculates performance statistics and
    verifies values based on the loaded data.

    Parameters
    ----------
    file_path : str
        The path to the file from which clusterer results should be loaded. The file
        should be a tsml formatted clusterer results file.
    calculate_stats : bool, default=True
        Whether to calculate performance statistics from the loaded results.
    verify_values : bool, default=True
        If the function should perform verification of the loaded values.

    Returns
    -------
    cr : ClustererResults
        A ClustererResults object containing the results loaded from the file.
    """
    with open(file_path) as file:
        lines = file.readlines()

        line1 = lines[0].split(",")
        line3 = lines[2].split(",")
        cl_acc = float(line3[0])
        n_clusters = int(line3[6])
        n_cases = len(lines) - 3

        line_size = len(lines[3].split(","))

        class_labels = np.zeros(n_cases)
        cluster = np.zeros(n_cases)
        probabilities = np.zeros((n_cases, n_clusters))

        if line_size > 4 + n_clusters:
            pred_times = np.zeros(n_cases)
        else:
            pred_times = None

        if line_size > 6 + n_clusters:
            pred_descriptions = []
        else:
            pred_descriptions = None

        for i in range(0, n_cases):
            line = lines[i + 3].split(",")
            class_labels[i] = int(line[0])
            cluster[i] = int(line[1])

            for j in range(0, n_clusters):
                probabilities[i, j] = float(line[3 + j])

            if pred_times is not None:
                pred_times[i] = float(line[4 + n_clusters])

            if pred_descriptions is not None:
                pred_descriptions.append(",".join(line[6 + n_clusters :]).strip())

    cr = ClustererResults(
        dataset_name=line1[0],
        clusterer_name=line1[1],
        split=line1[2],
        resample_id=None if line1[3] == "None" else int(line1[3]),
        time_unit=line1[4].lower(),
        description=",".join(line1[5:]).strip(),
        parameters=lines[1].strip(),
        fit_time=float(line3[1]),
        predict_time=float(line3[2]),
        benchmark_time=float(line3[3]),
        memory_usage=float(line3[4]),
        n_classes=int(line3[5]),
        n_clusters=n_clusters,
        class_labels=class_labels,
        predictions=cluster,
        probabilities=probabilities,
        pred_times=pred_times,
        pred_descriptions=pred_descriptions,
    )

    if calculate_stats:
        cr.calculate_statistics()

    if verify_values:
        cr.infer_size(overwrite=True)
        assert cr.n_cases == n_cases
        assert cr.n_clusters == n_clusters

        assert_allclose(probabilities.sum(axis=1), 1, rtol=1e-6)

        if calculate_stats:
            assert cr.clustering_accuracy == cl_acc

    return cr



"""Files for bakeoff redux 2023 publication."""

__all__ = [
    "_set_bakeoff_classifier",
    "_run_experiment",
    "bakeoff_classifiers",
]

from tsml_eval.publications.y2023.tsc_bakeoff.run_experiments import _run_experiment
from tsml_eval.publications.y2023.tsc_bakeoff.set_bakeoff_classifier import (
    _set_bakeoff_classifier,
    bakeoff_classifiers,
)



"""Alignment path plotting utilities."""

from typing import Optional

import numpy as np
from aeon.distances import cost_matrix as compute_cost_matrix
from aeon.distances._distance import alignment_path, pairwise_distance
from aeon.utils.validation._dependencies import _check_soft_dependencies


def _path_mask(cost_matrix, path, ax, theme=None):  # pragma: no cover
    _check_soft_dependencies("matplotlib")

    import matplotlib.colors as colorplt

    if theme is None:
        theme = colorplt.LinearSegmentedColormap.from_list("", ["#c9cacb", "white"])

    plot_matrix = np.zeros_like(cost_matrix)
    max_size = max(cost_matrix.shape)
    for i in range(max_size):
        for j in range(max_size):
            if (i, j) in path:
                plot_matrix[i, j] = 1.0
            elif cost_matrix[i, j] == np.inf:
                plot_matrix[i, j] = 0.0
            else:
                plot_matrix[i, j] = 0.25

    for i in range(max_size):
        for j in range(max_size):
            c = cost_matrix[j, i]
            ax.text(i, j, str(round(c, 2)), va="center", ha="center", size=10)
            ax.text(i, j, str(round(c, 2)), va="center", ha="center", size=10)

    ax.matshow(plot_matrix, cmap=theme)


def _pairwise_path(x, y, metric):  # pragma: no cover
    pw_matrix = pairwise_distance(x, y, metric=metric)
    path = []
    for i in range(pw_matrix.shape[0]):
        for j in range(pw_matrix.shape[1]):
            if i == j:
                path.append((i, j))
    return path, pw_matrix.trace(), pw_matrix


def _plot_path(
    x: np.ndarray,
    y: np.ndarray,
    metric: str,
    dist_kwargs: Optional[dict] = None,
    title: str = "",
    plot_over_pw: bool = False,
):  # pragma: no cover
    _check_soft_dependencies("matplotlib")

    import matplotlib.pyplot as plt

    if dist_kwargs is None:
        dist_kwargs = {}
    try:
        path, dist = alignment_path(x, y, metric=metric, **dist_kwargs)
        cost_matrix = compute_cost_matrix(x, y, metric=metric, **dist_kwargs)

        if metric == "lcss":
            _path = []
            for tup in path:
                _path.append(tuple(x + 1 for x in tup))
            path = _path

        if plot_over_pw is True:
            if metric == "lcss":
                pw = pairwise_distance(x, y, metric="euclidean")
                cost_matrix = np.zeros_like(cost_matrix)
                cost_matrix[1:, 1:] = pw
            else:
                pw = pairwise_distance(x, y, metric="squared")
                cost_matrix = pw
    except NotImplementedError:
        path, dist, cost_matrix = _pairwise_path(x, y, metric)

    plt.figure(1, figsize=(8, 8))
    x_size = x.shape[0]

    # definitions for the axes
    left, bottom = 0.01, 0.1
    w_ts = h_ts = 0.2
    left_h = left + w_ts + 0.02
    width = height = 0.65
    bottom_h = bottom + height + 0.02

    rect_s_y = [left, bottom, w_ts, height]
    rect_gram = [left_h, bottom, width, height]
    rect_s_x = [left_h, bottom_h, width, h_ts]

    ax_gram = plt.axes(rect_gram)
    ax_s_x = plt.axes(rect_s_x)
    ax_s_y = plt.axes(rect_s_y)

    _path_mask(cost_matrix, path, ax_gram)
    ax_gram.axis("off")
    ax_gram.autoscale(False)
    # ax_gram.plot([j for (i, j) in path], [i for (i, j) in path], "w-",
    #              linewidth=3.)

    ax_s_x.plot(np.arange(x_size), y, "b-", linewidth=3.0, color="#818587")
    ax_s_x.axis("off")
    ax_s_x.set_xlim((0, x_size - 1))

    ax_s_y.plot(-x, np.arange(x_size), "b-", linewidth=3.0, color="#818587")
    ax_s_y.axis("off")
    ax_s_y.set_ylim((0, x_size - 1))

    ax_s_x.set_title(title, size=10)

    return plt


def _plot_alignment(
    x, y, metric, dist_kwargs: Optional[dict] = None, title: str = ""
):  # pragma: no cover
    _check_soft_dependencies("matplotlib")

    import matplotlib.pyplot as plt

    if dist_kwargs is None:
        dist_kwargs = {}
    try:
        path, dist = alignment_path(x, y, metric=metric, **dist_kwargs)
    except NotImplementedError:
        path, dist, cost_matrix = _pairwise_path(x, y, metric)

    plt.figure(1, figsize=(8, 8))

    plt.plot(x, "b-", color="black")
    plt.plot(y, "g-", color="black")

    for positions in path:
        try:
            plt.plot(
                [positions[0], positions[1]],
                [x[positions[0]], y[positions[1]]],
                "--",
                color="#818587",
            )
        except Exception:
            continue
    plt.title(title)

    plt.tight_layout()
    return plt



"""A tsml wrapper for sklearn clusterers."""

__author__ = ["MatthewMiddlehurst"]
__all__ = ["SklearnToTsmlClusterer"]

import numpy as np
from sklearn.base import ClusterMixin
from sklearn.utils.validation import check_is_fitted
from tsml.base import BaseTimeSeriesEstimator, _clone_estimator


class SklearnToTsmlClusterer(ClusterMixin, BaseTimeSeriesEstimator):
    """Wrapper for sklearn estimators to use the tsml base class."""

    def __init__(
        self,
        clusterer=None,
        pad_unequal=False,
        concatenate_channels=False,
        clone_estimator=True,
        random_state=None,
    ):
        self.clusterer = clusterer
        self.pad_unequal = pad_unequal
        self.concatenate_channels = concatenate_channels
        self.clone_estimator = clone_estimator
        self.random_state = random_state

        super().__init__()

    def fit(self, X, y=None):
        """Wrap fit."""
        if self.clusterer is None:
            raise ValueError("Clusterer not set")

        X = self._validate_data(X=X)
        X = self._convert_X(
            X,
            pad_unequal=self.pad_unequal,
            concatenate_channels=self.concatenate_channels,
        )

        self._clusterer = (
            _clone_estimator(self.clusterer, self.random_state)
            if self.clone_estimator
            else self.clusterer
        )
        self._clusterer.fit(X, y)

        self.labels_ = self._clusterer.labels_

        return self

    def predict(self, X) -> np.ndarray:
        """Wrap predict."""
        check_is_fitted(self)

        X = self._validate_data(X=X, reset=False)
        X = self._convert_X(
            X,
            pad_unequal=self.pad_unequal,
            concatenate_channels=self.concatenate_channels,
        )

        return self._clusterer.predict(X)

    def _more_tags(self):
        return {
            "X_types": ["2darray"],
            "equal_length_only": (
                False if self.pad_unequal or self.concatenate_channels else True
            ),
        }



"""Regression Experiments: code for experiments as an alternative to orchestration.

This file is configured for runs of the main method with command line arguments, or for
single debugging runs. Results are written in a standard tsml format.
"""

__author__ = ["MatthewMiddlehurst"]

import sys

from tsml_eval.experiments import load_and_run_forecasting_experiment
from tsml_eval.experiments.set_forecaster import get_forecaster_by_name
from tsml_eval.experiments.tests import _FORECASTER_RESULTS_PATH
from tsml_eval.testing.testing_utils import _TEST_DATA_PATH
from tsml_eval.utils.arguments import parse_args
from tsml_eval.utils.experiments import _results_present


def run_experiment(args, overwrite=False):
    """Mechanism for testing forecasters using a csv data format."""
    # cluster run (with args), this is fragile
    # don't run threaded jobs on the cluster unless you have reserved the whole node
    # and know what you are doing
    if args is not None and args.__len__() > 0:
        print("Input args = ", args)
        args = parse_args(args)

        # this is also checked in load_and_run, but doing a quick check here so can
        # print a message and make sure data is not loaded
        if not overwrite and _results_present(
            args.results_path,
            args.estimator_name,
            args.dataset_name,
            resample_id=args.resample_id,
            split="TEST",
        ):
            print("Ignoring, results already present")
        else:
            load_and_run_forecasting_experiment(
                args.data_path,
                args.results_path,
                args.dataset_name,
                get_forecaster_by_name(
                    args.estimator_name,
                    random_state=(
                        args.resample_id
                        if args.random_seed is None
                        else args.random_seed
                    ),
                    n_jobs=args.n_jobs,
                    **args.kwargs,
                ),
                forecaster_name=args.estimator_name,
                random_seed=(
                    args.resample_id if args.random_seed is None else args.random_seed
                ),
                write_attributes=args.write_attributes,
                att_max_shape=args.att_max_shape,
                benchmark_time=args.benchmark_time,
                overwrite=args.overwrite,
            )
    # local run (no args)
    else:
        # These are example parameters, change as required for local runs
        # Do not include paths to your local directories here in PRs
        # If threading is required, see the threaded version of this file
        data_path = _TEST_DATA_PATH
        results_path = _FORECASTER_RESULTS_PATH
        estimator_name = "NaiveForecaster"
        dataset_name = "ShampooSales"
        random_seed = 0
        n_jobs = 1
        write_attributes = True
        att_max_shape = 0
        benchmark_time = True
        overwrite = False
        kwargs = {}

        forecaster = get_forecaster_by_name(
            estimator_name,
            random_state=random_seed,
            n_jobs=n_jobs,
            **kwargs,
        )
        print(f"Local Run of {estimator_name} ({forecaster.__class__.__name__}).")

        load_and_run_forecasting_experiment(
            data_path,
            results_path,
            dataset_name,
            forecaster,
            forecaster_name=estimator_name,
            random_seed=random_seed,
            write_attributes=write_attributes,
            att_max_shape=att_max_shape,
            benchmark_time=benchmark_time,
            overwrite=overwrite,
        )


if __name__ == "__main__":
    """
    Example simple usage, with arguments input via script or hard coded for testing.
    """
    print("Running forecasting_experiments.py main")
    run_experiment(sys.argv[1:])



"""Tests for IterativeVotingClustering."""

import numpy as np

# import pytest
from aeon.datasets import load_arrow_head
from sklearn.metrics import rand_score

# from tsml_eval.estimators.clustering.consensus.ivc import IterativeVotingClustering
from tsml_eval.estimators.clustering.consensus.ivc_from_file import (
    FromFileIterativeVotingClustering,
)
from tsml_eval.testing.testing_utils import _TEST_RESULTS_PATH

# from tsml.utils.testing import generate_3d_test_data


# @pytest.mark.parametrize("init", ["plus", "random", "aligned"])
# def test_ivc_init_methods(init):
#     """Test IVC init methods."""
#     X, y = generate_3d_test_data(n_labels=4)
#
#     ivc = IterativeVotingClustering(
#         n_clusters=4,
#         max_iterations=10,
#         init=init,
#     )
#     ivc.fit(X, y)
#     preds = ivc.predict(X)
#
#     assert ivc.labels_.shape == (len(X),)
#     assert preds.shape == (len(X),)


def test_from_file_iterative_voting_clustering():
    """Test SimpleVote from file with ArrowHead results."""
    X_train, y_train = load_arrow_head(split="train")
    X_test, y_test = load_arrow_head(split="test")

    file_paths = [
        _TEST_RESULTS_PATH + "/clustering/PAM-DTW/Predictions/ArrowHead/",
        _TEST_RESULTS_PATH + "/clustering/PAM-ERP/Predictions/ArrowHead/",
        _TEST_RESULTS_PATH + "/clustering/PAM-MSM/Predictions/ArrowHead/",
    ]

    ivc = FromFileIterativeVotingClustering(
        clusterers=file_paths, n_clusters=3, max_iterations=100, random_state=0
    )
    ivc.fit(X_train, y_train)
    preds = ivc.predict(X_test)

    assert ivc.labels_.shape == (len(X_train),)
    assert isinstance(ivc.labels_, np.ndarray)
    assert rand_score(y_train, ivc.labels_) >= 0.6
    assert preds.shape == (len(X_test),)
    assert isinstance(preds, np.ndarray)
    assert rand_score(y_test, preds) >= 0.6



"""Experiments tests."""

__all__ = [
    "_CLASSIFIER_RESULTS_PATH",
    "_CLUSTERER_RESULTS_PATH",
    "_FORECASTER_RESULTS_PATH",
    "_REGRESSOR_RESULTS_PATH",
]

from tsml_eval.testing.testing_utils import _TEST_OUTPUT_PATH

_CLASSIFIER_RESULTS_PATH = _TEST_OUTPUT_PATH + "/classification/"

_CLUSTERER_RESULTS_PATH = _TEST_OUTPUT_PATH + "/clustering/"

_FORECASTER_RESULTS_PATH = _TEST_OUTPUT_PATH + "/forecasting/"

_REGRESSOR_RESULTS_PATH = _TEST_OUTPUT_PATH + "/regression/"



"""Test publication utilities."""

import os

from tsml_eval.testing.testing_utils import _TEST_EVAL_PATH, _TEST_OUTPUT_PATH
from tsml_eval.utils.publications import extract_publication_csv_from_evaluation


def test_extract_publication_csv_from_evaluation():
    """Test extracting publication results CSVs from evaluation directories."""
    extract_publication_csv_from_evaluation(
        "Accuracy",
        f"{_TEST_EVAL_PATH}/classification/",
        f"{_TEST_OUTPUT_PATH}/eval_result_files_test/",
    )

    assert os.path.exists(
        f"{_TEST_EVAL_PATH}/classification/Accuracy/all_resamples/1NN-DTW_accuracy.csv"
    )
    assert os.path.exists(
        f"{_TEST_EVAL_PATH}/classification/Accuracy/all_resamples/ROCKET_accuracy.csv"
    )
    assert os.path.exists(
        f"{_TEST_EVAL_PATH}/classification/Accuracy/all_resamples/TSF_accuracy.csv"
    )
    assert os.path.exists(
        f"{_TEST_OUTPUT_PATH}/eval_result_files_test/1NN-DTW_accuracy.csv"
    )
    assert os.path.exists(
        f"{_TEST_OUTPUT_PATH}/eval_result_files_test/ROCKET_accuracy.csv"
    )
    assert os.path.exists(
        f"{_TEST_OUTPUT_PATH}/eval_result_files_test/TSF_accuracy.csv"
    )

    os.remove(f"{_TEST_OUTPUT_PATH}/eval_result_files_test/1NN-DTW_accuracy.csv")
    os.remove(f"{_TEST_OUTPUT_PATH}/eval_result_files_test/ROCKET_accuracy.csv")
    os.remove(f"{_TEST_OUTPUT_PATH}/eval_result_files_test/TSF_accuracy.csv")



"""Utility functions fordata resampling."""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]
__all__ = [
    "resample_data",
    "resample_data_indices",
    "stratified_resample_data",
    "stratified_resample_data_indices",
]


import numpy as np
from sklearn.utils import check_random_state


def resample_data(X_train, y_train, X_test, y_test, random_state=None):
    """Resample data without replacement using a random state.

    Reproducible resampling. Combines train and test, randomly resamples, then returns
    new train and test.

    Parameters
    ----------
    X_train : np.ndarray or list of np.ndarray
        Train data in a 2d or 3d ndarray or list of arrays.
    y_train : np.ndarray
        Train data labels.
    X_test : np.ndarray or list of np.ndarray
        Test data in a 2d or 3d ndarray or list of arrays.
    y_test : np.ndarray
        Test data labels.
    random_state : int, RandomState instance or None, default=None
        If `int`, random_state is the seed used by the random number generator;
        If `RandomState` instance, random_state is the random number generator;
        If `None`, the random number generator is the `RandomState` instance used
        by `np.random`.

    Returns
    -------
    train_X : np.ndarray or list of np.ndarray
        New train data.
    train_y : np.ndarray
        New train labels.
    test_X : np.ndarray or list of np.ndarray
        New test data.
    test_y : np.ndarray
        New test labels.
    """
    if isinstance(X_train, np.ndarray):
        is_array = True
    elif isinstance(X_train, list):
        is_array = False
    else:
        raise ValueError(
            "X_train must be a np.ndarray array or list of np.ndarray arrays"
        )

    # add both train and test to a single dataset
    all_labels = np.concatenate((y_train, y_test), axis=None)
    all_data = (
        np.concatenate([X_train, X_test], axis=0) if is_array else X_train + X_test
    )

    # shuffle data indices
    rng = check_random_state(random_state)
    indices = np.arange(len(all_data), dtype=int)
    rng.shuffle(indices)

    train_indices = indices[: len(X_train)]
    test_indices = indices[len(X_train) :]

    # split the shuffled data into train and test
    X_train = (
        all_data[train_indices] if is_array else [all_data[i] for i in train_indices]
    )
    y_train = all_labels[train_indices]
    X_test = all_data[test_indices] if is_array else [all_data[i] for i in test_indices]
    y_test = all_labels[test_indices]

    return X_train, y_train, X_test, y_test


def resample_data_indices(y_train, y_test, random_state=None):
    """Return data resample indices without replacement using a random state.

    Reproducible resampling. Combines train and test, randomly resamples, then returns
    the new position for both the train and test set. Uses indices for a combined train
    and test set, with test indices appearing after train indices.

    Parameters
    ----------
    y_train : np.ndarray
        Train data labels.
    y_test : np.ndarray
        Test data labels.
    random_state : int, RandomState instance or None, default=None
        If `int`, random_state is the seed used by the random number generator;
        If `RandomState` instance, random_state is the random number generator;
        If `None`, the random number generator is the `RandomState` instance used
        by `np.random`.

    Returns
    -------
    train_indices : np.ndarray
        The index of cases to use in the train set from the combined train and test
        data.
    test_indices : np.ndarray
        The index of cases to use in the test set from the combined train and test data.
    """
    # shuffle data indices
    rng = check_random_state(random_state)
    indices = np.arange(len(y_train) + len(y_test), dtype=int)
    rng.shuffle(indices)

    train_indices = indices[: len(y_train)]
    test_indices = indices[len(y_train) :]

    return train_indices, test_indices


def stratified_resample_data(X_train, y_train, X_test, y_test, random_state=None):
    """Stratified resample data without replacement using a random state.

    Reproducible resampling. Combines train and test, resamples to get the same class
    distribution, then returns new train and test.

    Parameters
    ----------
    X_train : np.ndarray or list of np.ndarray
        Train data in a 2d or 3d ndarray or list of arrays.
    y_train : np.ndarray
        Train data labels.
    X_test : np.ndarray or list of np.ndarray
        Test data in a 2d or 3d ndarray or list of arrays.
    y_test : np.ndarray
        Test data labels.
    random_state : int, RandomState instance or None, default=None
        If `int`, random_state is the seed used by the random number generator;
        If `RandomState` instance, random_state is the random number generator;
        If `None`, the random number generator is the `RandomState` instance used
        by `np.random`.

    Returns
    -------
    train_X : np.ndarray or list of np.ndarray
        New train data.
    train_y : np.ndarray
        New train labels.
    test_X : np.ndarray or list of np.ndarray
        New test data.
    test_y : np.ndarray
        New test labels.
    """
    if isinstance(X_train, np.ndarray):
        is_array = True
    elif isinstance(X_train, list):
        is_array = False
    else:
        raise ValueError(
            "X_train must be a np.ndarray array or list of np.ndarray arrays"
        )

    # add both train and test to a single dataset
    all_labels = np.concatenate((y_train, y_test), axis=None)
    all_data = (
        np.concatenate([X_train, X_test], axis=0) if is_array else X_train + X_test
    )

    # shuffle data indices
    rng = check_random_state(random_state)

    # count class occurrences
    unique_train, counts_train = np.unique(y_train, return_counts=True)
    unique_test, counts_test = np.unique(y_test, return_counts=True)

    # ensure same classes exist in both train and test
    assert list(unique_train) == list(unique_test)

    if is_array:
        shape = list(X_train.shape)
        shape[0] = 0

    X_train = np.zeros(shape) if is_array else []
    y_train = np.zeros(0)
    X_test = np.zeros(shape) if is_array else []
    y_test = np.zeros(0)

    # for each class
    for label_index in range(len(unique_train)):
        # get the indices of all instances with this class label and shuffle them
        label = unique_train[label_index]
        indices = np.where(all_labels == label)[0]
        rng.shuffle(indices)

        train_indices = indices[: counts_train[label_index]]
        test_indices = indices[counts_train[label_index] :]

        # extract data from corresponding indices
        train_cases = (
            all_data[train_indices]
            if is_array
            else [all_data[i] for i in train_indices]
        )
        train_labels = all_labels[train_indices]
        test_cases = (
            all_data[test_indices] if is_array else [all_data[i] for i in test_indices]
        )
        test_labels = all_labels[test_indices]

        # concat onto current data from previous loop iterations
        X_train = (
            np.concatenate([X_train, train_cases], axis=0)
            if is_array
            else X_train + train_cases
        )
        y_train = np.concatenate([y_train, train_labels], axis=None)
        X_test = (
            np.concatenate([X_test, test_cases], axis=0)
            if is_array
            else X_test + test_cases
        )
        y_test = np.concatenate([y_test, test_labels], axis=None)

    return X_train, y_train, X_test, y_test


def stratified_resample_data_indices(y_train, y_test, random_state=None):
    """Return stratified data resample indices without replacement using a random state.

    Reproducible resampling. Combines train and test, resamples to get the same class
    distribution, then returns the new position for both the train and test set.
    Uses indices for a combined train and test set, with test indices appearing after
    train indices.

    Parameters
    ----------
    y_train : np.ndarray
        Train data labels.
    y_test : np.ndarray
        Test data labels.
    random_state : int, RandomState instance or None, default=None
        If `int`, random_state is the seed used by the random number generator;
        If `RandomState` instance, random_state is the random number generator;
        If `None`, the random number generator is the `RandomState` instance used
        by `np.random`.

    Returns
    -------
    train_indices : np.ndarray
        The index of cases to use in the train set from the combined train and test
        data.
    test_indices : np.ndarray
        The index of cases to use in the test set from the combined train and test data.
    """
    # add both train and test to a single dataset
    all_labels = np.concatenate((y_train, y_test), axis=None)

    # shuffle data indices
    rng = check_random_state(random_state)

    # count class occurrences
    unique_train, counts_train = np.unique(y_train, return_counts=True)
    unique_test, counts_test = np.unique(y_test, return_counts=True)

    # ensure same classes exist in both train and test
    assert list(unique_train) == list(unique_test)

    train_indices = np.zeros(0, dtype=int)
    test_indices = np.zeros(0, dtype=int)

    # for each class
    for label_index in range(len(unique_train)):
        # get the indices of all instances with this class label and shuffle them
        label = unique_train[label_index]
        indices = np.where(all_labels == label)[0]
        rng.shuffle(indices)

        train_indices = np.concatenate(
            [train_indices, indices[: counts_train[label_index]]], axis=None
        )
        test_indices = np.concatenate(
            [test_indices, indices[counts_train[label_index] :]], axis=None
        )

    return train_indices, test_indices



"""Tests for consensus clustering algorithms."""

from sklearn.utils.estimator_checks import parametrize_with_checks

from tsml_eval.estimators.clustering.consensus.ivc import IterativeVotingClustering
from tsml_eval.estimators.clustering.consensus.simple_vote import SimpleVote


@parametrize_with_checks(
    [
        SimpleVote(n_clusters=2),
        IterativeVotingClustering(n_clusters=2, max_iterations=5),
    ]
)
def test_sklearn_checks(estimator, check):
    """Test consensus clusterers with sklearn estimator checks."""
    check(estimator)



from math import floor

import numpy as np
from sklearn.utils.validation import check_random_state
from aeon.distances import (
    euclidean_pairwise_distance,
    dtw_pairwise_distance,
    ddtw_pairwise_distance,
    wdtw_pairwise_distance,
    wddtw_pairwise_distance,
    erp_pairwise_distance,
    lcss_pairwise_distance,
    twe_pairwise_distance,
    msm_pairwise_distance
)

DISTANCE_CANDIDATES = [
    "euclidean",
    "dtw",
    "ddtw",
    "dtw-r",
    "ddtw-r",
    "wdtw",
    "wddtw",
    "erp",
    "lcss",
    "twe",
    "msm",
]


class DistanceSplitter:
    """EE-based splitter for TS-CHIEF implementation."""

    @staticmethod
    def generate(X, y, random_state=None):
        """Generate a randomized distance splitter candidate."""
        _, dims, length = X.shape
        rng = check_random_state(random_state)

        splitter = DistanceSplitter()
        splitter.dim = rng.randint(dims)

        metric = rng.choice(DISTANCE_CANDIDATES)
        example = X[0, splitter.dim, :]
        splitter.metric = metric
        if metric == "euclidean":
            pass
        elif metric == "dtw":
            pass
        elif metric == "ddtw":
            pass
        elif metric == "dtw-r":
            max_warp = floor((length + 1) / 4)
            warp = rng.randint(0, max_warp + 1)
            splitter.window = warp / length
        elif metric == "ddtw-r":
            max_warp = floor((length + 1) / 4)
            warp = rng.randint(0, max_warp + 1)
            splitter.window = warp / length
        elif metric == "wdtw":
            splitter.g = rng.uniform(0, 1)
        elif metric == "wddtw":
            splitter.g = rng.uniform(0, 1)
        elif metric == "erp":
            sigma = np.std(X[:, splitter.dim, :])
            splitter.g = rng.uniform(sigma / 5, sigma)
        elif metric == "lcss":
            sigma = np.std(X[:, splitter.dim, :])
            splitter.epsilon = rng.uniform(sigma / 5, sigma)
            max_warp = floor((length + 1) / 4)
            splitter.window = rng.randint(0, max_warp + 1)
        elif metric == "twe":
            splitter.nu = rng.choice(
                [0.00001, 0.0001, 0.0005, 0.001, 0.005, 0.01, 0.05, 0.1, 0.5, 1]
            )
            splitter.lmbda = rng.choice(np.arange(0, 10) / 9)
        elif metric == "msm":
            splitter.c = rng.choice(
                [
                    0.01,
                    0.01375,
                    0.0175,
                    0.02125,
                    0.025,
                    0.02875,
                    0.0325,
                    0.03625,
                    0.04,
                    0.04375,
                    0.0475,
                    0.05125,
                    0.055,
                    0.05875,
                    0.0625,
                    0.06625,
                    0.07,
                    0.07375,
                    0.0775,
                    0.08125,
                    0.085,
                    0.08875,
                    0.0925,
                    0.09625,
                    0.1,
                    0.136,
                    0.172,
                    0.208,
                    0.244,
                    0.28,
                    0.316,
                    0.352,
                    0.388,
                    0.424,
                    0.46,
                    0.496,
                    0.532,
                    0.568,
                    0.604,
                    0.64,
                    0.676,
                    0.712,
                    0.748,
                    0.784,
                    0.82,
                    0.856,
                    0.892,
                    0.928,
                    0.964,
                    1,
                    1.36,
                    1.72,
                    2.08,
                    2.44,
                    2.8,
                    3.16,
                    3.52,
                    3.88,
                    4.24,
                    4.6,
                    4.96,
                    5.32,
                    5.68,
                    6.04,
                    6.4,
                    6.76,
                    7.12,
                    7.48,
                    7.84,
                    8.2,
                    8.56,
                    8.92,
                    9.28,
                    9.64,
                    10,
                    13.6,
                    17.2,
                    20.8,
                    24.4,
                    28,
                    31.6,
                    35.2,
                    38.8,
                    42.4,
                    46,
                    49.6,
                    53.2,
                    56.8,
                    60.4,
                    64,
                    67.6,
                    71.2,
                    74.8,
                    78.4,
                    82,
                    85.6,
                    89.2,
                    92.8,
                    96.4,
                    100,
                ]
            )

        splitter.exemplars = []
        classes = np.unique(y)
        for c in classes:
            group = np.argwhere(y == c).ravel()
            exemplar_idx = rng.choice(group)
            splitter.exemplars.append(X[exemplar_idx, splitter.dim, :])
        splitter.exemplars = np.array(splitter.exemplars)

        return splitter

    def split(self, X):
        """Split incoming data."""
        samples, _, _ = X.shape
        X = X[:, self.dim, :]
        y = self.exemplars

        if self.metric == "euclidean":
            pairwise_distances = euclidean_pairwise_distance(X, y)
        elif self.metric == "dtw":
            pairwise_distances = dtw_pairwise_distance(X, y)
        elif self.metric == "ddtw":
            pairwise_distances = ddtw_pairwise_distance(X, y)
        elif self.metric == "dtw-r":
            pairwise_distances = dtw_pairwise_distance(X, y, self.window)
        elif self.metric == "ddtw-r":
            pairwise_distances = dtw_pairwise_distance(X, y, self.window)
        elif self.metric == "wdtw":
            pairwise_distances = wdtw_pairwise_distance(X, y, g=self.g)
        elif self.metric == "wddtw":
            pairwise_distances = wddtw_pairwise_distance(X, y, g=self.g)
        elif self.metric == "erp":
            pairwise_distances = erp_pairwise_distance(X, y, g=self.g)
        elif self.metric == "lcss":
            pairwise_distances = lcss_pairwise_distance(X, y, window=self.window, epsilon=self.epsilon)
        elif self.metric == "twe":
            pairwise_distances = twe_pairwise_distance(X, y, lmbda=self.lmbda, nu=self.nu)
        else:
            pairwise_distances = msm_pairwise_distance(X, y, c=self.c)

        split_idx = np.argmin(pairwise_distances, axis=1)

        return split_idx



"""Working area for summarising results.

1. Stages in validating and collating results, using tsml tools to generate all
summary stats and extracting stats from these for the website

2. Using existing results on website for new comparison.

"""

from tsml_eval.evaluation.multiple_estimator_evaluation import \
     evaluate_clusterers_by_problem, evaluate_classifiers_by_problem, evaluate_regressors_by_problem
from aeon.datasets.tsc_data_lists import univariate_equal_length
from aeon.benchmarking import plot_critical_difference
from aeon.benchmarking.results_loaders import get_estimator_results

import pandas as pd
import numpy as np

import numpy as np

from aeon.classification.tests.test_classification_reproduction import _print_results_for_classifier
from aeon.transformations.tests.test_transformer_reproduction import _print_results_for_transformer


from aeon.datasets.tsc_data_lists import univariate_equal_length
import os
import glob

def count_files(directory, pattern):
    """Count the number of files in a directory matching a pattern."""
    path = os.path.join(directory, pattern)
    return len(glob.glob(path))

def count_lines(full_path):
    line_count = 0
    with open(full_path) as file:
        for line in file:
            line_count += 1
    return line_count

def count_complete(root, datasets, algorithms, resamples =30):
    """Count the number of results files in each directory."""
    table=[]
    # Function to count files matching the pattern in a directory
    for d in datasets:
        count =0
        counts = np.zeros(len(algorithms))
        for r in algorithms:
            # Creating a table of counts

            directory = f"{root}\\{r}\\Predictions\\{d}"
            counts[count]= count_files(directory, "TestResample*.csv")
            count=count+1
        table.append(counts)
    df = pd.DataFrame(counts, index =tser_new, columns=algorithms)
    # Compare each element to expected resample
    n_resamples = df >= resamples
    # Sum up the True values in each column
    counts = n_resamples.sum()
    return counts


def collate_by_algorithm(estimators, measures, destination, location):
    """Extract the individual files for all estimators/measures combinations."""
    ## Set up destination files
    if not os.path.exists(destination):
        os.makedirs(destination)
    for m in measures:
        m= m.lower()
        path = destination+"\\"+m
        if not os.path.exists(path):
            os.makedirs(path)
        for e in estimators:
            source = location + "\\" + m+"\\all_resamples\\"+e+"_"+m+".csv"
            dest = destination + "\\" + m+"\\"+e+"_"+m+".csv"
            shutil.copy(source,dest)



print(univariate_equal_length)
print(list(univariate_equal_length))

from aeon.benchmarking import get_estimator_results, get_estimator_results_as_array
est = ["STC","RDST","ROCKET","MrSQM"]
import pandas as pd

new_res = pd.read_csv("C:\\Temp\\Accuracy_mean.csv")
data_names = new_res.iloc[:,0].tolist()
cls_names = new_res.columns[1:].tolist()

print(cls_names)

new_results = new_res.iloc[:,3].values.tolist()

array = np.zeros((len(data_names),len(est)+1))
published_results = get_estimator_results(estimators=est, default_only=False)
for i in range(len(data_names)):
    for j in range(len(est)):
        array[i][j] =np.average(published_results[est[j]][data_names[i]])
    array[i][len(est)] = new_results[i]
from aeon.benchmarking import plot_critical_difference
est.append(cls_names[2])
cd = plot_critical_difference(array, est)
cd.show()

print(est)

print(np.mean(array,axis=0))


def compare_to_reference(new_path, est):
    """Compare new results to reference results."""
    new_res = pd.read_csv(new_path)
    data_names = new_res.iloc[:,0].tolist()
    cls_names = new_res.columns[1:].tolist()
    new_results = new_res.iloc[:,1:].values
    array = np.zeros((len(data_names),len(est)+len(cls_names)))
    published_results = get_estimator_results(estimators=est, default_only=False)
    for i in range(len(data_names)):
     for j in range(len(est)):
         array[i,j] =np.average(published_results[est[j]][data_names[i]])
     for j in range(len(cls_names)):
          array[i,len(est)+j] = new_results[i,j]

    for i in range(len(cls_names)):
     est.append(cls_names[i])

    cd = plot_critical_difference(array, est)
    cd.show()


if __name__ == "__main__":
    print("AFC")



"""Classifiers used in the publication."""

__author__ = ["MatthewMiddlehurst"]

import numpy as np

from tsml_eval.utils.functions import str_in_nested_list

rist_classifiers = [
    ["FreshPRINCEClassifier", "freshprince"],
    ["ShapeletTransformClassifier", "stc", "stc-2hour"],
    ["RDSTClassifier", "rdst"],
    ["RSTSFClassifier", "rstsf", "r-stsf"],
    ["drcif", "DrCIFClassifier"],
    ["RocketClassifier", "rocket"],
    ["HIVECOTEV2", "hc2"],
    ["RISTClassifier", "rist", "rist-extrat"],
    "rist-rf",
    "rist-ridgecv",
    ["RandomIntervalClassifier", "intervalpipeline", "i-pipeline"],
]


def _set_rist_classifier(
    classifier_name,
    random_state=None,
    n_jobs=1,
    **kwargs,
):
    c = classifier_name.lower()

    if not str_in_nested_list(rist_classifiers, c):
        raise ValueError(f"UNKNOWN CLASSIFIER: {c} in _set_rist_classifier")

    if c == "freshprinceclassifier" or c == "freshprince":
        from aeon.classification.feature_based import FreshPRINCEClassifier

        return FreshPRINCEClassifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "shapelettransformclassifier" or c == "stc" or c == "stc-2hour":
        from aeon.classification.shapelet_based import ShapeletTransformClassifier

        return ShapeletTransformClassifier(
            transform_limit_in_minutes=120,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "rdstclassifier" or c == "rdst":
        from tsml.shapelet_based import RDSTClassifier

        return RDSTClassifier(random_state=random_state)
    elif c == c == "rstsfclassifier" or c == "rstsf" or c == "r-stsf":
        from tsml.interval_based import RSTSFClassifier

        return RSTSFClassifier(
            n_estimators=500, random_state=random_state, n_jobs=n_jobs, **kwargs
        )
    elif c == "drcif" or c == "drcifclassifier":
        from aeon.classification.interval_based import DrCIFClassifier

        return DrCIFClassifier(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif c == "rocketclassifier" or c == "rocket":
        from aeon.classification.convolution_based import RocketClassifier

        return RocketClassifier(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "hivecotev2" or c == "hc2":
        from aeon.classification.hybrid import HIVECOTEV2

        return HIVECOTEV2(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif c == "ristclassifier" or c == "rist" or c == "rist-extrat":
        from sklearn.ensemble import ExtraTreesClassifier
        from tsml.hybrid import RISTClassifier

        return RISTClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=ExtraTreesClassifier(n_estimators=500, criterion="entropy"),
            **kwargs,
        )
    elif c == "rist-rf":
        from sklearn.ensemble import RandomForestClassifier
        from tsml.hybrid import RISTClassifier

        return RISTClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=RandomForestClassifier(n_estimators=500, criterion="entropy"),
            **kwargs,
        )
    elif c == "rist-ridgecv":
        from sklearn.linear_model import RidgeClassifierCV
        from sklearn.pipeline import make_pipeline
        from sklearn.preprocessing import StandardScaler
        from tsml.hybrid import RISTClassifier

        return RISTClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=make_pipeline(
                StandardScaler(with_mean=False),
                RidgeClassifierCV(alphas=np.logspace(-4, 4, 20)),
            ),
            **kwargs,
        )
    elif (
        c == "randomintervalclassifier" or c == "intervalpipeline" or c == "i-pipeline"
    ):
        from sklearn.ensemble import ExtraTreesClassifier
        from tsml.interval_based import RandomIntervalClassifier
        from tsml.transformations import (
            ARCoefficientTransformer,
            Catch22Transformer,
            FunctionTransformer,
            PeriodogramTransformer,
        )
        from tsml.utils.numba_functions.general import first_order_differences_3d
        from tsml.utils.numba_functions.stats import (
            row_iqr,
            row_mean,
            row_median,
            row_numba_max,
            row_numba_min,
            row_ppv,
            row_slope,
            row_std,
        )

        def sqrt_times_15_plus_5_mv(X):
            return int(
                np.sqrt(X.shape[2]) * np.sqrt(X.shape[1]) * 15 + 5
            )  # pragma: no cover

        interval_features = [
            Catch22Transformer(outlier_norm=True, replace_nans=True),
            row_mean,
            row_std,
            row_slope,
            row_median,
            row_iqr,
            row_numba_min,
            row_numba_max,
            row_ppv,
        ]
        series_transformers = [
            None,
            FunctionTransformer(func=first_order_differences_3d, validate=False),
            PeriodogramTransformer(use_pyfftw=True),
            ARCoefficientTransformer(replace_nan=True),
        ]

        return RandomIntervalClassifier(
            random_state=random_state,
            n_jobs=n_jobs,
            n_intervals=sqrt_times_15_plus_5_mv,
            features=interval_features,
            series_transformers=series_transformers,
            estimator=ExtraTreesClassifier(n_estimators=500, criterion="entropy"),
            **kwargs,
        )



"""Experiment runner for RIST pipeline publication."""

__author__ = ["MatthewMiddlehurst"]

import os
import sys

os.environ["MKL_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["NUMEXPR_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["OMP_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["TF_NUM_INTEROP_THREADS"] = "1"
os.environ["TF_NUM_INTRAOP_THREADS"] = "1"

from tsml.base import _clone_estimator

from tsml_eval.experiments import load_and_run_regression_experiment
from tsml_eval.publications.y2023.rist_pipeline.set_rist_regressor import (
    _set_rist_regressor,
)
from tsml_eval.publications.y2023.rist_pipeline.tests import _RIST_TEST_RESULTS_PATH
from tsml_eval.testing.testing_utils import _TEST_DATA_PATH
from tsml_eval.utils.arguments import parse_args
from tsml_eval.utils.experiments import _results_present

regressors = [
    "InceptionTime",
    "ROCKET",
    "DrCIF",
    "FreshPRINCE",
    "RDST",
    "RIST-ExtraT",
    "RIST-RF",
    "RIST-RidgeCV",
]


def _run_regression_experiment(args):
    if args is None or args.__len__() < 1:
        data_path = _TEST_DATA_PATH
        results_path = _RIST_TEST_RESULTS_PATH
        regressor = "ROCKET"
        dataset_name = "MinimalGasPrices"
        resample_id = 0
        n_jobs = 1
        kwargs = {}
        overwrite = False
    else:
        print("Input args = ", args)
        args = parse_args(args)
        data_path = args.data_path
        results_path = args.results_path
        regressor = args.estimator_name
        dataset_name = args.dataset_name
        resample_id = args.resample_id
        n_jobs = args.n_jobs
        kwargs = args.kwargs
        overwrite = args.overwrite

    # Skip if not overwrite and results already present
    # this is also checked in load_and_run, but doing a quick check here so can
    # print a message and make sure data is not loaded
    if not overwrite and _results_present(
        results_path,
        regressor,
        dataset_name,
        resample_id=resample_id,
        split="TEST",
    ):
        print("Ignoring, results already present")
    else:
        load_and_run_regression_experiment(
            data_path,
            results_path,
            dataset_name,
            (
                _set_rist_regressor(
                    regressor,
                    random_state=resample_id,
                    n_jobs=n_jobs,
                    **kwargs,
                )
                if isinstance(regressor, str)
                else _clone_estimator(regressor, resample_id)
            ),
            regressor_name=regressor,
            resample_id=resample_id,
            overwrite=overwrite,
        )


if __name__ == "__main__":
    """
    Example simple usage, with arguments input via script or hard coded for testing.

    1. Edit the arguments from line 48-55 to suit your experiment, The most important
       are:
         data_path: the path to the data
         results_path: the path to the results
         regressor: the name of the regressor to use (check set_rist_regressor.py),
         or an estimator object
         resample_id: the data resample id and random seed to use
    2. Run the script, if the experiment runs successfully a set of folders and a
       results csv file will be created in the results path.

    For evaluation of the written results, you can use the evaluation package, see
    our examples for usage:
    https://github.com/time-series-machine-learning/tsml-eval/blob/main/examples/

    For using your own regressor, any regressor following the sklearn, aeon,
    or tsml interface should be compatible with this file.
    """
    print("Running run_regression_experiments.py main")
    _run_regression_experiment(sys.argv[1:])
